# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/15_arc.ipynb.

# %% auto 0
__all__ = ['ARCFitnessFactory', 'BaseARCFitness', 'SumSquareOfDiff', 'AverageMaxOfDiff', 'Euclidean', 'ARCDataProcessor',
           'ARCEnv']

# %% ../nbs/15_arc.ipynb 3
import os, math
import numpy as np
import gym
from time import sleep
from abc import ABC, abstractmethod

# %% ../nbs/15_arc.ipynb 4
from .putils import limit_large_float, get_abs_tol
from .helpers import ChallengesDataManager

# %% ../nbs/15_arc.ipynb 7
class ARCFitnessFactory:
    factories = {}
    def addFactory(id, functionFactory):
        ARCFitnessFactory.factories.put[id] = functionFactory
    addFactory = staticmethod(addFactory)
    # A Template Method:
    def createFunction(id):
        if not ARCFitnessFactory.factories.__contains__(id):
            ARCFitnessFactory.factories[id] = \
              eval(id + '.Factory()')
        return ARCFitnessFactory.factories[id].create()
    createFunction = staticmethod(createFunction)


# %% ../nbs/15_arc.ipynb 9
class BaseARCFitness(ABC):
    "Base class of an ARCFitness. This class is not used directly by developers, but defines the functionality common to all."

    def get_dim_metric(self, output_array=None, env_array=None, control_set=None):   
        dim_metric = 0
        if 'dims' in control_set and len(control_set) == 1:
            dim_metric = (env_array.shape[0] - output_array.shape[0]) ** 2 + (env_array.shape[1] - output_array.shape[1]) ** 2

        return dim_metric

# %% ../nbs/15_arc.ipynb 11
class SumSquareOfDiff(BaseARCFitness):
    "A function that sums the square of the diffrences of two arrays."
    
    def __call__(self, output_array=None, env_array=None, control_set=None):
        # First metric: square of the difference between the dimensions
        dim_metric = self.get_dim_metric(output_array, env_array, control_set)

        # Second metric: square of the difference between each element in the arrays
        element_metric = 0
        if 'cells' in control_set:
            diff = env_array[:output_array.shape[0], :output_array.shape[1]] - output_array
            diff = np.where(np.isnan(diff), 0, diff)
            diff = np.where(np.isnan(env_array[:output_array.shape[0], :output_array.shape[1]]), 10, diff)
            element_metric = np.sum(diff ** 2)

        # Final metric
        if 'dims' in control_set and len(control_set) == 1:
            final_metric = dim_metric
        elif 'cells' in control_set and len(control_set) == 1:
            final_metric = element_metric
        else:
            final_metric = dim_metric + element_metric

        return final_metric

    class Factory:
        def create(self): return SumSquareOfDiff()     
        

# %% ../nbs/15_arc.ipynb 13
class AverageMaxOfDiff(BaseARCFitness):
    "A function that sums the square of the diffrences of two arrays."
    
    def __call__(self, output_array=None, env_array=None, control_set=None):
        # First metric: square of the difference between the dimensions
        dim_metric = self.get_dim_metric(output_array, env_array, control_set)

        # Second metric: square of the difference between each element in the arrays
        element_metric = 0
        if 'cells' in control_set:
            diff = env_array[:output_array.shape[0], :output_array.shape[1]] - output_array
            diff = np.where(np.isnan(diff), 0, diff)
            diff = np.where(np.isnan(env_array[:output_array.shape[0], :output_array.shape[1]]), 10, diff)
            abs_diff = np.abs(diff)
            maximum = np.max(abs_diff)
            average = np.average(abs_diff)
            element_metric = max(average, maximum)

        # Final metric
        if 'dims' in control_set and len(control_set) == 1:
            final_metric = dim_metric
        elif 'cells' in control_set and len(control_set) == 1:
            final_metric = element_metric
        else:
            final_metric = dim_metric + element_metric

        return final_metric

    class Factory:
        def create(self): return AverageMaxOfDiff()     

# %% ../nbs/15_arc.ipynb 15
class Euclidean(BaseARCFitness):
    "A function that sums the square of the diffrences of two arrays."
    
    def __call__(self, output_array=None, env_array=None, control_set=None):
        # First metric: square of the difference between the dimensions
        dim_metric = self.get_dim_metric(output_array, env_array, control_set)

        # Second metric: square of the difference between each element in the arrays
        element_metric = 0
        if 'cells' in control_set:
            diff = env_array[:output_array.shape[0], :output_array.shape[1]] - output_array
            diff = np.where(np.isnan(diff), 0, diff)
            diff = np.where(np.isnan(env_array[:output_array.shape[0], :output_array.shape[1]]), 10, diff)
            element_metric = np.sqrt(np.sum(diff ** 2))

            abs_diff = np.abs(diff)
            maximum = np.max(abs_diff)
            if maximum < 0.5:
                element_metric = maximum

        # Final metric
        if 'dims' in control_set and len(control_set) == 1:
            final_metric = dim_metric
        elif 'cells' in control_set and len(control_set) == 1:
            final_metric = element_metric
        else:
            final_metric = dim_metric + element_metric

        return final_metric

    class Factory:
        def create(self): return Euclidean()     

# %% ../nbs/15_arc.ipynb 18
class ARCDataProcessor:
    def __init__(self, config_dict, arc_dict):

        self.fitness_function_type = ARCFitnessFactory.createFunction(config_dict.get('fitness', 'SumSquareOfDiff'))

        self.arc_dict = arc_dict
        self.grid_shape = self.determine_grid_shape()
        
        if 'grid_shape' in config_dict:
            self.grid_shape = config_dict.get('grid_shape')
        
        self.input_set = config_dict.get('input_set', ['env'])
        if not all(item in {'env', 'inputs', 'outputs'} for item in self.input_set):
            raise ValueError("input_set must be a list containing 'env', 'inputs', and/or 'outputs'")
        
        self.control_set = config_dict.get('control_set', ['default'])
        if not all(item in {'dims', 'cells'} for item in self.control_set):
            raise ValueError("control_set must be a list containing 'dims' and/or 'cells'")
        
        if 'env' not in self.input_set:
            self.input_set.append('env')

        self.index = config_dict.get('index', 0)
        self.initial_index = self.index if 'index' in config_dict else None
        self.dataset = config_dict.get('dataset', None)

        if self.dataset is None:
            raise ValueError("Dataset must be defined in environment properties as either 'train' or 'test'.")

        if 'dims' in self.control_set and self.grid_shape is None:
            raise ValueError("grid_shape cannot be None when 'dims' is in control_set")

        if self.dataset == 'test':
            self.test_output_array = np.array(config_dict['test_output_array'])
        else:
            self.test_output_array = None

        self.create_env()
        self.info = self.create_info()

    def add_rows(self, num_rows):
        num_rows = round(num_rows)
        if num_rows == 0:
            return
        self.env = np.pad(self.env, ((0, num_rows), (0, 0)), mode='constant', constant_values=0)

    def remove_rows(self, num_rows):
        num_rows = round(num_rows)
        if num_rows > self.env.shape[0]:
            num_rows = self.env.shape[0] - 1
        if num_rows > 0 and self.env.shape[0] > num_rows:
            self.env = self.env[:-num_rows, :]

    def add_columns(self, num_columns):
        num_columns = round(num_columns)
        if num_columns == 0:
            return
        self.env = np.pad(self.env, ((0, 0), (0, num_columns)), mode='constant', constant_values=0)

    def remove_columns(self, num_columns):
        num_columns = round(num_columns)
        if num_columns > self.env.shape[1]:
            num_columns = self.env.shape[1] - 1
        if num_columns > 0 and self.env.shape[1] > num_columns:
            self.env = self.env[:, :-num_columns]

    def process_remaining_values(self, values):
        values = np.clip(values, -9, 9)
        if len(values) != self.env.size:
            raise ValueError(f"The number of elements in values ({len(values)}) must equal the number of elements in self.env ({self.env.size})")
        for i, value in enumerate(values):
            row, col = divmod(i, self.env.shape[1])
            if row < self.env.shape[0] and col < self.env.shape[1]:
                self.env[row, col] = np.clip(self.env[row, col] + value, 0, 9)

    def get_array(self, key, index, sub_key):
        return np.array(self.arc_dict[key][index][sub_key])

    def next(self):
        self.index += 1
        if self.index >= len(self.arc_dict[self.dataset]):
            return False
        if self.initial_index is not None:
            return False
        # print(f"Index: {self.index}")
        self.create_env()
        return True

    def reset(self):
        if self.initial_index is None:
            self.index = 0
        else:
            self.index = self.initial_index
        self.create_env()

    def create_env(self):
        self.env = np.array(self.arc_dict[self.dataset][self.index]['input'], dtype=np.float32)

    def process_dimensions(self, actions):

        if len(actions) == 2:
            num_rows, num_cols = actions
        elif len(actions) == 1:
            num_rows = num_cols = actions[0]
        else:
            raise ValueError("Actions must have one or two values")

        num_rows = limit_large_float(num_rows, 100)
        num_cols = limit_large_float(num_cols, 100)

        if num_rows > 0:
            self.add_rows(num_rows)
        elif num_rows < 0:
            self.remove_rows(abs(num_rows))

        if num_cols > 0:
            self.add_columns(num_cols)
        elif num_cols < 0:
            self.remove_columns(abs(num_cols))

    def get_input_dimensions(self):
        return np.array(self.arc_dict[self.dataset][self.index]['input']).shape

    def get_output_dimensions(self):
        if self.dataset == 'test':
            if self.test_output_array is None:
                raise ValueError("test_output_array is not defined")
            return self.test_output_array.shape

        return np.array(self.arc_dict[self.dataset][self.index]['output']).shape

    def get_env_dimensions(self):
        return self.env.shape

    def apply_actions(self, actions):
        if len(actions) != self.info['num_actions']:
            raise Exception(f'Number of actions from hierarchy {len(actions)} should equal expected actions {self.info["num_actions"]}.')
        value_index = 0
        if 'dims' in self.control_set:
            if self.grid_shape == 'equal':
                self.process_dimensions(actions[:1])
                value_index = 1
            elif self.grid_shape == 'unequal':
                self.process_dimensions(actions[:2])
                value_index = 2
        
        if 'cells' in self.control_set:
            self.process_remaining_values(actions[value_index:])

    def create_info(self):
        info = {}
        info['num_actions'] = 0

        if 'dims' in self.control_set:
            info['grid_shape'] = self.grid_shape

            if self.grid_shape == 'equal':
                info['num_actions'] = 1
                info['dims'] = len(self.input_set)
            elif self.grid_shape == 'unequal':
                info['num_actions'] = 2
                info['dims'] = len(self.input_set) * 2

        if 'cells' in self.control_set:
            if 'env' in self.input_set:
                edims = self.get_env_dimensions()
                info['env'] = edims
                info['num_actions'] += edims[0] * edims[1]
            
            if 'inputs' in self.input_set:
                idims = self.get_input_dimensions()
                info['inputs'] = idims
                # info['num_actions'] += idims[0] * idims[1]

            if 'outputs' in self.input_set:
                odims = self.get_output_dimensions()
                info['outputs'] = odims
                # info['num_actions'] += odims[0] * odims[1]

        return info

    def set_test_output_array(self, array):
        self.test_output_array = array

    def get_output(self, dataset):
        if dataset == 'test':
            if self.test_output_array is None:
                raise ValueError("test_output_array is not defined")
            return self.test_output_array
        return np.array(self.arc_dict[dataset][self.index]['output'])

    def fitness_function_arrays(self, output_array, env_array):


        return self.fitness_function_type(output_array = output_array, env_array = env_array, control_set = self.control_set) 


        # # First metric: square of the difference between the dimensions
        # dim_metric = 0
        # if 'dims' in self.control_set and len(self.control_set) == 1:
        #     dim_metric = (env_array.shape[0] - output_array.shape[0]) ** 2 + (env_array.shape[1] - output_array.shape[1]) ** 2

        # # Second metric: square of the difference between each element in the arrays
        # element_metric = 0
        # if 'cells' in self.control_set:
        #     diff = env_array[:output_array.shape[0], :output_array.shape[1]] - output_array
        #     diff = np.where(np.isnan(diff), 0, diff)
        #     diff = np.where(np.isnan(env_array[:output_array.shape[0], :output_array.shape[1]]), 10, diff)
        #     element_metric = np.sum(diff ** 2)

        # # Final metric
        # if 'dims' in self.control_set and len(self.control_set) == 1:
        #     final_metric = dim_metric
        # elif 'cells' in self.control_set and len(self.control_set) == 1:
        #     final_metric = element_metric
        # else:
        #     final_metric = dim_metric + element_metric

        # return final_metric

    def fitness_function(self):
        output_array = self.get_output(self.dataset)
        env_array = self.env
        return self.fitness_function_arrays(output_array, env_array)

    def get_input(self, dataset):
        return np.array(self.arc_dict[dataset][self.index]['input'])

    def get_num_elements(self, dataset, array_key):
        return np.array(self.arc_dict[dataset][self.index][array_key]).size

    def get_info(self):
        return self.info

    def determine_grid_shape(self):
        output_dims = [np.array(task['output']).shape for task in self.arc_dict['train']]
        input_dims = [np.array(task['input']).shape for task in self.arc_dict['train']]

        if len(set(output_dims)) == 1:
            if len(set(input_dims)) == 1 and output_dims[0] == input_dims[0]:
                return None
            return 'equal'
        return 'unequal'

    def get_env_inputs_names(self):
        input_names = []
        if 'dims' in self.control_set:
            dfactor = 2 if self.grid_shape == 'unequal' else 1
            if 'env' in self.input_set:
                input_names.append('IWE')
                if dfactor == 2:
                    input_names.append('IHE')
            if 'inputs' in self.input_set:
                input_names.append('IWI')
                if dfactor == 2:
                    input_names.append('IHI')
            if 'outputs' in self.input_set:
                input_names.append('IW0')
                if dfactor == 2:
                    input_names.append('IHO')
            
        if 'env' in self.info:
            dims = self.info['env']
            num = dims[0] * dims[1]
            for i in range(num):
                input_names.append(f'IE{i+1:03}')
            
        if 'inputs' in self.info:
            dims = self.info['inputs']
            num = dims[0] * dims[1]
            for i in range(num):
                input_names.append(f'II{i+1:03}')

        if 'outputs' in self.info:
            dims = self.info['outputs']
            num = dims[0] * dims[1]
            for i in range(num):
                input_names.append(f'IO{i+1:03}')
            
        return input_names

    def get_env_inputs_indexes(self):
        ninputs = 0

        if 'dims' in self.info:
            ninputs += self.info['dims']

        if 'env' in self.info:
            ninputs += self.info['env'][0] * self.info['env'][1]

        if 'inputs' in self.info:
            ninputs += self.info['inputs'][0] * self.info['inputs'][1]

        if 'outputs' in self.info:
            ninputs += self.info['outputs'][0] * self.info['outputs'][1]

        env_inputs_indexes = [i for i in range(ninputs)]

        return env_inputs_indexes

    def get_env_array(self):
        return self.env

    def get_state(self):
        state = {'inputs': {}}
        self.info['num_actions'] = 0

        def set_state_dimensions(dic, state_key, dimensions):
            if state_key in self.input_set:
                dic[state_key] = dimensions
                # self.info['num_actions'] += 1

        if 'dims' in self.control_set:
            dims = {}
            if self.grid_shape == 'equal':
                self.info['num_actions'] = 1
                set_state_dimensions(dims, 'env', (self.get_env_dimensions()[1],))
                set_state_dimensions(dims, 'inputs', (self.get_input_dimensions()[1],))
                set_state_dimensions(dims, 'outputs', (self.get_output_dimensions()[1],))
            elif self.grid_shape == 'unequal':
                self.info['num_actions'] = 2  # increment twice for unequal grid shape
                set_state_dimensions(dims, 'env', self.get_env_dimensions())
                set_state_dimensions(dims, 'inputs', self.get_input_dimensions())
                set_state_dimensions(dims, 'outputs', self.get_output_dimensions())
            state['inputs']['dims'] = dims

        if 'cells' in self.control_set:
            cells = {}
            if 'env' in self.input_set:
                cells['env'] = self.env
                dims = self.get_env_dimensions()
                self.info['env'] = dims
                self.info['num_actions'] += dims[0] * dims[1]

            if 'inputs' in self.input_set:
                cells['inputs'] = self.get_array(self.dataset, self.index, 'input')

            if 'outputs' in self.input_set:
                cells['outputs'] = self.get_array(self.dataset, self.index, 'output')
            state['inputs']['cells'] = cells

        return state, self.info



# %% ../nbs/15_arc.ipynb 23
class ARCEnv(gym.Env):
    def __init__(self, namespace=""):
        super(ARCEnv, self).__init__()
        self.debug = False
        self.index = 0
        self.env = None
        self.dimensions = []
        self._fitness = 10000  # Initialize fitness to 10000
        self.state = []
        self.done = False

        # Class variables
        self.dataset = None
        self.namespace = namespace
        self.fitness_list = []
        self.gradient_list = []
        # Render settings
        self.screen_width = 1000
        self.screen_height = 500
        self.cell_size = 30
        self.left_pad = 20
        self.height_pad = 20
        self.grid_down = 50
        self.symbol_down = 150
        self.screen = None
        self.isopen = True

        from matplotlib import colors       

        self.cmap = colors.ListedColormap(['#000000', '#0074D9', '#FF4136', '#2ECC40', '#FFDC00',
                                           '#AAAAAA', '#F012BE', '#FF851B', '#7FDBFF', '#870C25'])
        self.norm = colors.Normalize(vmin=0, vmax=9)

        self.fitness_label_font_size = 20
        self.fitness_value_font_size = 50

        self.iteration = 1  # Initialize iteration to 1
        self.code = ""  # Example property, ensure to set self.code elsewhere in your class
        self.data = None  # Placeholder for data
        self.arc_data = None  # Placeholder for ARCDataProcessor
        self.num_actions = 0  # Initialize num_actions

    def initialise(self, properties, arc_dict):
        """
        Initialize the environment with properties and arc_dict.
        """
        self.dataset = properties.get('dataset', None)
        self.code = properties.get('code', "")
        self.arc_data = ARCDataProcessor(properties, arc_dict)
        self.runs = properties.get('runs', len(arc_dict['train'])) if 'index' in properties else properties.get('runs', 1) / len(arc_dict['train'])
        self.abs_tol_ARC_resolved = properties['tolerances'].get('atARCresolved', 0.01)
        self.abs_tol_ARC_display = properties['tolerances'].get('atARCdisplay', 0.5)
        

        self.reset()

    def step(self, actions):
        """
        Take a step in the environment.
        """
        self.arc_data.apply_actions(actions)

        # Always calculate fitness regardless of dataset type
        self.fitness = self.arc_data.fitness_function()

        self.state, self.info = self.arc_data.get_state()
        
        self.iteration += 1  # Increment iteration
        if self.iteration > self.runs:
            if self.debug:
                print('step', self.iteration, self.index)

            self.done = True
        return self.state, self.fitness, self.done, self.info

    def reset(self):
        """
        Reset the environment to the initial state.
        """
        self.arc_data.reset()

        # Always calculate fitness
        self.fitness = self.arc_data.fitness_function()

        self.done = False
        self.state, self.info = self.arc_data.get_state()
        self.iteration = 1  # Reset iteration
        self.num_actions = self.info['num_actions']  # Set num_actions
        self.fitness_list = []  # Initialize an empty list for fitness
        self.gradient_list = []  # Initialize an empty list for gradient
        self.fitness_isclose_to_zero = False

    def next(self):
        """
        Move to the next state in arc_dict.
        Returns False if the current fitness is not close to zero.
        """
        # self.fitness_list.append(self.fitness)
        if self.debug:
            print('next', self.iteration, self.fitness_list)
        
        self.iteration = 1  # Reset iteration to 1
        self.done = False
        
        # uncomment the following lines to check if fitness is close to zero discontinue processing the next index in the environment        
        # if not self.fitness_isclose_to_zero:
        #     return False
        return self.arc_data.next()

    def add_to_fitness_list(self, fitness): 
        """
        Add the provided fitness to the fitness list.
        """
        self.fitness_list.append(fitness)

    def add_to_gradient_list(self, gradient): 
        """
        Add the provided gradient to the gradient list.
        """
        self.gradient_list.append(gradient) 


    def is_environment_resolved(self):
        max_fitness = max(self.fitness_list)
        return max_fitness < self.abs_tol_ARC_resolved  


    def get_environment_score(self):
        """
        Get the environment score.
        """
        if len(self.fitness_list) == 0:
            return self.fitness
        return max(self.fitness_list)


    def get_num_actions(self):
        """
        Get the number of actions.
        """
        return self.num_actions

    def get_env_inputs_names(self):
        """
        Get the environment input names.
        """
        return self.arc_data.get_env_inputs_names()

    def get_env_inputs_indexes(self):
        """
        Get the environment input indexes.
        """
        return self.arc_data.get_env_inputs_indexes()

    def set_dataset(self, dataset):
        """
        Set the dataset and update the value in arc_data.
        """
        self.dataset = dataset
        self.arc_data.set_dataset(dataset)

    def call_fitness_function_arrays(self):
        """
        Call fitness_function_arrays on arc_data.
        """
        return self.arc_data.fitness_function_arrays()

    def get_env_array(self):
        """
        Get the environment array from arc_data.
        """
        return self.arc_data.get_env_array()

    @property
    def fitness(self):
        """
        Get the current fitness value.
        """
        return self._fitness

    @fitness.setter
    def fitness(self, value):
        """
        Set the current fitness value.
        """
        self._fitness = value

    def render(self, mode='human'):
        """
        Render the environment using Pygame.
        """
        import pygame

        def draw_grid(screen, grid, top_left_x, top_left_y, cell_size):
            for i, row in enumerate(grid):
                for j, value in enumerate(row):
                    if top_left_x + j * cell_size < self.screen_width and top_left_y + i * cell_size < self.screen_height:
                        rvalue = round(value)
                        color = self.cmap(self.norm(rvalue))[:3]  # Get RGB only, excluding alpha
                        color = tuple(int(c * 255) for c in color)  # Multiply each element by 255
                        pygame.draw.rect(screen, color, (top_left_x + j * cell_size, top_left_y + i * cell_size, cell_size, cell_size))
                        pygame.draw.rect(screen, (255, 255, 255), (top_left_x + j * cell_size, top_left_y + i * cell_size, cell_size, cell_size), 1)
            # Draw a black line around the grid
            if top_left_x < self.screen_width and top_left_y < self.screen_height:
                pygame.draw.rect(screen, (0, 0, 0), (top_left_x, top_left_y, cell_size * len(grid[0]), cell_size * len(grid)), 2)

        if self.screen is None:
            pygame.init()
            self.screen = pygame.display.set_mode((self.screen_width, self.screen_height))
            pygame.display.set_caption('ARC Environment')

        input_grid = self.arc_data.get_input(self.dataset)
        output_grid = self.arc_data.get_output(self.dataset)
        env_grid = self.arc_data.env

        # Top left coordinates
        input_grid_x = self.left_pad
        input_grid_y = self.grid_down
        arrow_img_x = input_grid_x + len(input_grid[0]) * self.cell_size + self.left_pad
        arrow_img_y = self.symbol_down
        output_grid_x = arrow_img_x + 50 + self.left_pad
        output_grid_y = self.grid_down
        equals_img_x = output_grid_x + len(output_grid[0]) * self.cell_size + self.left_pad
        equals_img_y = self.symbol_down
        env_grid_x = equals_img_x + 50 + self.left_pad
        env_grid_y = self.grid_down
        fitness_text_x = env_grid_x + env_grid.shape[1] * self.cell_size + self.left_pad
        fitness_text_y = self.grid_down + self.height_pad
        fitness_value_y = fitness_text_y + self.fitness_label_font_size + self.height_pad
        tick_cross_y = fitness_value_y + self.fitness_value_font_size + self.height_pad
        table_y = self.grid_down + max(len(input_grid), len(output_grid), env_grid.shape[0]) * self.cell_size + self.height_pad

        # Adjust screen size if necessary
        if fitness_text_x + self.fitness_value_font_size > self.screen_width:
            self.screen_width = fitness_text_x + self.fitness_value_font_size + self.left_pad
            self.screen = pygame.display.set_mode((self.screen_width, self.screen_height))
        if env_grid_y + env_grid.shape[0] * self.cell_size > self.screen_height:
            self.screen_height = env_grid_y + env_grid.shape[0] * self.cell_size + self.height_pad
            self.screen = pygame.display.set_mode((self.screen_width, self.screen_height))

        self.screen.fill((255, 255, 255))  # Clear the screen

        draw_grid(self.screen, input_grid, input_grid_x, input_grid_y, self.cell_size)
        draw_grid(self.screen, output_grid, output_grid_x, output_grid_y, self.cell_size)
        draw_grid(self.screen, env_grid, env_grid_x, env_grid_y, self.cell_size)

        # Load and scale images
        arrow_img = pygame.transform.scale(pygame.image.load('images/arrow.png'), (50, 50))
        equals_img = pygame.transform.scale(pygame.image.load('images/equals.jpg'), (50, 50))
        green_tick_img = pygame.transform.scale(pygame.image.load('images/green_tick.png'), (50, 50))
        red_cross_img = pygame.transform.scale(pygame.image.load('images/red-cross.png'), (50, 50))

        if arrow_img_x < self.screen_width and arrow_img_y < self.screen_height:
            self.screen.blit(arrow_img, (arrow_img_x, arrow_img_y))
        if equals_img_x < self.screen_width and equals_img_y < self.screen_height:
            self.screen.blit(equals_img, (equals_img_x, equals_img_y))

        # Display fitness text
        label_font = pygame.font.Font(None, self.fitness_label_font_size)
        value_font = pygame.font.Font(None, self.fitness_value_font_size)
        label_font.set_bold(True)
        fitness_label = label_font.render("Fitness:", True, (0, 0, 0))
        fitness_value = value_font.render(f"{self.fitness:.2f}", True, (0, 0, 0))
        if fitness_text_x < self.screen_width and fitness_text_y < self.screen_height:
            self.screen.blit(fitness_label, (fitness_text_x, fitness_text_y))
        if fitness_text_x < self.screen_width and fitness_value_y < self.screen_height:
            self.screen.blit(fitness_value, (fitness_text_x, fitness_value_y))

        # Change fitness test to use math.isclose
        if math.isclose(self.fitness, 0, abs_tol=self.abs_tol_ARC_display):
            if fitness_text_x < self.screen_width and tick_cross_y < self.screen_height:
                self.screen.blit(green_tick_img, (fitness_text_x, tick_cross_y))
        else:
            if fitness_text_x < self.screen_width and tick_cross_y < self.screen_height:
                self.screen.blit(red_cross_img, (fitness_text_x, tick_cross_y))

        # Draw table with Code, Iteration, Index, and Dataset
        table_font = pygame.font.Font(None, 24)
        table_labels = ["Code", "Iteration", "Index", "Dataset"]
        table_values = [os.path.splitext(self.code)[0], self.iteration, self.arc_data.index, self.dataset]  # Display code without file extension
        table_x = self.left_pad
        table_y = table_y

        for i, (label, value) in enumerate(zip(table_labels, table_values)):
            label_surface = table_font.render(label, True, (0, 0, 0))
            value_surface = table_font.render(str(value), True, (0, 0, 0))
            self.screen.blit(label_surface, (table_x + i * 150, table_y))
            self.screen.blit(value_surface, (table_x + i * 150, table_y + 30))

        pygame.display.flip()    
        pygame.event.pump()  # Handle Pygame events to keep the window responsive

    def close(self):
        """
        Close the environment, save the screen to an HTML and an image file.
        """
        import pygame
        if self.screen is not None:
            os.makedirs("c:/tmp/arc/", exist_ok=True)
            image_filename = f"c:/tmp/arc/screen_image_{self.namespace}.png"
            html_filename = f"c:/tmp/arc/screen_image_{self.namespace}.html"
            
            pygame.image.save(self.screen, image_filename)
            
            # Save the screen image to an HTML format
            with open(html_filename, "w") as f:
                f.write(f"<html><body><img src='screen_image_{self.namespace}.png'></body></html>")

            pygame.display.quit()
            pygame.quit()
            self.isopen = False





# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/20_experiments.ipynb.

# %% auto 0
__all__ = ['CometExperimentManager']

# %% ../nbs/20_experiments.ipynb 4
import os, json, csv
from comet_ml import API, APIExperiment, start
from .hierarchy import PCTHierarchy  
from comet_ml.query import Metric

# %% ../nbs/20_experiments.ipynb 5
class CometExperimentManager:
    def __init__(self, api_key: str = None, workspace: str = None):
        self.api = API(api_key)
        self.workspace = workspace

    def get_all_artifacts_indexed(self):
        """Retrieve all artifacts and sort them by source experiment key."""
        filename = '/tmp/artifacts/artifacts_results.json'
        
        # Create directory if it doesn't exist
        os.makedirs(os.path.dirname(filename), exist_ok=True)
        
        if os.path.exists(filename):
            with open(filename, 'r') as file:
                return json.load(file)
            
        artifacts = self.api.get_artifact_list(workspace=self.workspace)
        experiment = start(workspace=self.workspace)
        results = {}
        for artifact_dict in artifacts['artifacts']:
            id = artifact_dict['name']
            try:
                logged_artifact = experiment.get_artifact(id)
                print(logged_artifact.source_experiment_key, id)
                results[logged_artifact.source_experiment_key] = id
            except Exception as e:
                print(f"Error retrieving artifact {id}: {e}")

        # Save results to a file
        with open(filename, "w") as file:
            json.dump(results, file, indent=4)
        experiment.end()
        return results

    def get_experiments_by_metrics(self, project_name: str = None, score_threshold: float = None, reward_threshold: float = None, max : bool = False):
        """
        Retrieve experiments for a project where the metric 'score' is less than
        score_threshold and 'reward' is greater than or equal to reward_threshold.
        """
        if max:
            # experiments = self.api.query(self.workspace, project_name, Metric("score") < score_threshold & Metric("reward") >= reward_threshold)
            experiments = self.api.query(self.workspace, project_name, Metric("score") > score_threshold )
        else:
        # experiments = self.api.query(self.workspace, project_name, Metric("score") < score_threshold & Metric("reward") >= reward_threshold)
            experiments = self.api.query(self.workspace, project_name, Metric("score") < score_threshold )
 
        return experiments

    def get_artifact_name(self, experiment: APIExperiment = None):
        """Retrieve the name of an artifact from an experiment."""
        artifacts = experiment.get_artifacts()
        return artifacts[0]['artifact_name'] if artifacts else None

    def download_and_run_artifact(self, artifact_name: str = None, seeds: list[int] = None):
        """
        Download an artifact to '/tmp/artifacts/' and run PCTHierarchy.run_from_file
        with the artifact filename, returning the score value for each run.
        """
        download_path = f"/tmp/artifacts/"
        os.makedirs(os.path.dirname(download_path), exist_ok=True)
 
        full_path = os.path.join(download_path, artifact_name)
        if os.path.exists(full_path):
            pass
        else:
            experiment = start(workspace=self.workspace)
            logged_artifact  = experiment.get_artifact(artifact_name)
            # print(logged_artifact.source_experiment_key)
            # local_artifact = logged_artifact.download(download_path)
            logged_artifact.download(download_path)
            # filename = f"{local_artifact.download_local_path}{artifact_name}"
            experiment.end()
    
        rewards = []
        for seed in seeds:
            hierarchy, score = PCTHierarchy.run_from_file(full_path, seed=seed)
            metrics = hierarchy.get_environment().get_metrics()
            # print(f'Score={score:0.3f} {metrics}')
            rewards.append(metrics['reward'])



        return rewards

    def get_original_metrics(self, experiment: APIExperiment = None):
        """Retrieve the metrics for an experiment."""
        metrics = {}
        metrics['score'] = eval(experiment.get_metrics("score")[0]['metricValue'])
        hyperparameters = experiment.get_parameters_summary()

        for param in hyperparameters:
            if param['name'] == 'mode':
                metrics['mode'] = param['valueCurrent']
                break

        metrics['name'] = experiment.name
        return metrics

    def run_experiments_and_record_results(self, project_name: str = None, experiments: list[APIExperiment] = None, artifact_results: dict = None, num_runs: int = 0, output_csv: str = None):
        """
        Run each experiment for a given project a specified number of times and record the score and the
        number of times the reward is 100, -100, or something else. Save results to a CSV file.
        """
        results = []
        for experiment in experiments:
            try:
                artifact_name = artifact_results[experiment.id]
            except KeyError:
                print(f"WARNING: Artifact not found for experiment {experiment.id}")
                continue

            metrics = self.get_original_metrics(experiment)

            print(f"Running experiment {experiment.id} in project {project_name} with artifact {artifact_name}")
            if not artifact_name:
                continue

            rewards = self.download_and_run_artifact(artifact_name, seeds=range(num_runs))
            reward_counts = {'100': 0, '-100': 0, 'other': 0}

            for reward in rewards:
                if reward == 100:
                    reward_counts['100'] += 1
                elif reward == -100:
                    reward_counts['-100'] += 1
                else:
                    reward_counts['other'] += 1
            print(f"Rewards: {reward_counts}")
            results.append({
                'name': metrics['name'],
                'score': round(metrics['score'], 5),
                'mode': metrics['mode'],
                'reward_100': reward_counts['100'],
                'reward_-100': reward_counts['-100'],
                'reward_other': reward_counts['other'],
                'experiment_key': '/'.join(("https://www.comet.com", self.workspace, project_name, experiment.id)),
                'artifact_name': artifact_name
            })
        # Sort results by 'reward_100' in descending order, then by 'reward_other' in descending order
        results.sort(key=lambda x: (-x['reward_100'], -x['reward_other']))
        # Save results to CSV
        with open(output_csv, mode='w', newline='') as csvfile:
            writer = csv.DictWriter(csvfile, fieldnames=['name', 'score', 'mode', 'reward_100', 'reward_-100', 'reward_other', 'experiment_key', 'artifact_name'])
            writer.writeheader()
            writer.writerows(results)

    def get_workspace_projects(self):
        """
        Get all projects from the current workspace.
        
        Returns:
            list: A list of project names in the workspace
        """
        api = API()
        projects = api.get_projects(workspace=self.workspace)
        return projects
        # return [project['name'] for project in projects]



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/10_webots.ipynb.

# %% auto 0
__all__ = ['WebotsHelper']

# %% ../nbs/10_webots.ipynb 3
class WebotsHelper():
    
    def __init__(self, name=None, mode=1):
        self.name = name
        self.mode = mode
        if self.mode==1:
            self.sensor_names = [ 'LHipPitch', 'LKneePitch', 'LAnklePitch', 'RHipPitch', 'RKneePitch', 'RAnklePitch']
            self.num_sensors=len(self.sensor_names)
            self.num_links=self.num_sensors

    def get_sensor_indexes(self):
        if self.mode==1:
            return [i for i in range(self.num_sensors)]
        
        
    def get_sensor_names(self):
        return self.sensor_names

    def get_references(self):
        if self.mode==1:
            return [0.5]#[i for i in range(self.num_sensors)]
        
    def get_sensor_values(self, msg):            
        return list(msg.values())
    
    def get_num_links(self):
        return self.num_links
        
    def get_actions_dict(self, inputs):
        actions = {}
        for i in range(len(inputs)):
            actions[self.sensor_names[i]]=round(inputs[i], 3)
                               
        return actions



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/18_applications.ipynb.

# %% auto 0
__all__ = ['run_wind_turbine']

# %% ../nbs/18_applications.ipynb 3
from .environment_processing import EnvironmentProcessingFactory


# %% ../nbs/18_applications.ipynb 4
def run_wind_turbine(target, drive, filepath):
    verbosed = {'debug': 0, 'hpct_verbose':0}

    # drive = get_gdrive()
    # root_path=get_root_path()
    root_path=""
    results_props = {'comparisons' : True, 'comparisons_print_plots': False}
    env_name = 'WindTurbine'
    args = {'file': filepath, 'env_name':env_name, 'verbosed':verbosed, 'drive':drive, 'root_path':root_path, 'max' : True} | results_props
    # args = {'file': filepath, 'env_name':env_name, 'verbosed':verbosed, 'max' : True} | results_props

    environment_properties={'series': target, 'zero_threshold': 1, 'reward_type': 'power', 'keep_history': True, 'range': 'test'}

    env_proc = EnvironmentProcessingFactory.createEnvironmentProcessing(f'{env_name}EnvironmentProcessing')
    env_proc.set_properties(args=args)

    env_proc.get_experiment()
    rtn = env_proc.results(filepath=filepath, environment_properties=environment_properties)





# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/19_pctexamples.ipynb.

# %% auto 0
__all__ = ['PCTExamples']

# %% ../nbs/19_pctexamples.ipynb 3
from .hierarchy import PCTHierarchy
from .environment_processing import EnvironmentProcessingFactory

# %% ../nbs/19_pctexamples.ipynb 5
class PCTExamples:
    """
    PCTExamples class provides methods to load a PCT hierarchy from a configuration file, 
    summarize the hierarchy, get the configuration, draw the hierarchy, run the hierarchy 
    for a specified number of steps, and close the environment.
    Attributes:
        hierarchy (PCTHierarchy): The loaded PCT hierarchy.
        env (Environment): The environment associated with the PCT hierarchy.
    """

    def __init__(self, config_file, min=True, early_termination=False, history=False, additional_props=None, render = False):
        """
        Initializes the PCTExamples instance by loading the hierarchy from the given configuration file.
        """
        self.config_file = config_file
        self.hierarchy, self.env, self.environment_properties = PCTHierarchy.load_from_file(config_file, min=min, early_termination=early_termination, 
                                                                                            history=history, additional_props=additional_props, render=render)
        self.history_data = None
        self.history = history

    def summary(self):
        """
        Prints a summary of the hierarchy.
        """
        self.hierarchy.summary()

    def get_config(self):
        """
        Returns the configuration of the hierarchy.
        """
        return self.hierarchy.get_config()

    def draw(self, with_labels=True, with_edge_labels=False, font_size=12, font_weight='bold', font_color='black', 
                       color_mapping={'PL':'aqua','OL':'limegreen','CL':'goldenrod', 'RL':'red', 'I':'silver', 'A':'yellow'},
                       node_size=500, arrowsize=25, align='horizontal', file=None, figsize=(8,8), move={}, draw_fig=True,
                       node_color=None, layout={'r':2,'c':1,'p':2, 'o':0}, funcdata=False, interactive_mode=False, experiment=None):
        """
        Draws the hierarchy with the specified parameters and returns the figure.
        """
        fig = self.hierarchy.draw(with_labels, with_edge_labels, font_size, font_weight, font_color, color_mapping, node_size, 
                            arrowsize, align, file, figsize, move, draw_fig, node_color, layout, funcdata, interactive_mode, experiment)
        
        return fig

    def run(self, steps=1, verbose=False):#, render=False):
        """
        Runs the hierarchy for the specified number of steps and returns the result.
        """
        # self.hierarchy.get_preprocessor()[0].set_render(render)
        return self.hierarchy.run(steps, verbose)

    def results(self, verbose=False, args=None, environment_properties=None):
        print(self.environment_properties)
        env_name = self.env.get_name()
        env_proc = EnvironmentProcessingFactory.createEnvironmentProcessing(f'{env_name}EnvironmentProcessing')
        allargs={'drive':'', 'verbosed':{'hpct_verbose':verbose}, 'max': False} | args
        env_proc.set_properties(allargs)
        env_proc.results(filepath=self.config_file, environment_properties=self.environment_properties|environment_properties, hierarchy=self.hierarchy)


    def close(self):
        """
        Closes the environment.
        """
        self.env.close()

    def get_history_keys(self):
        """
        Returns the keys of the history data of the hierarchy.
        """
        if self.history :
            if self.history_data is None:
                self.history_data = self.hierarchy.get_history_data()
            return list(self.history_data.keys())
        else:
            print("History is not enabled")

        return None

    def set_history_data(self):
        """
        Sets the history data of the hierarchy.
        """
        if self.history :

            if self.history_data is None:
                self.history_data = self.hierarchy.get_history_data()

            return self.history_data
        else:
            print("History is not enabled")

        return None

    def get_config_file_contents(self):
        """
        Reads and returns the contents of the config file.
        """
        with open(self.config_file, 'r') as file:
            return file.read()

    def plot_history(self, plots=None, title_prefix='', plots_dir=None, plots_figsize=(12, 6), history_data=None):
        """
        Plots the history of the hierarchy.
        """
        print(plots)
        plots = self.hierarchy.get_plots_config(plots, title_prefix)
        print(plots)

        from os import sep
        for plot in plots:
            print(plot)
            plotfile=None
            if plots_dir:
                plotfile = plots_dir + sep + plot['title'] + '-' + str(self.hierarchy.get_namespace()) + '.png'
            fig = self.hierarchy.hierarchy_plots(title=plot['title'], plot_items=plot['plot_items'], figsize=plots_figsize, file=plotfile, history=history_data)
            # import matplotlib.pyplot as plt
            # plt.close(fig)  # Close the figure here
    
        return fig
    
    def plot_single(self, plot=None, title_prefix='', plots_dir=None, plots_figsize=(12, 6), history_data=None):
        """
        Plot one item of the history of the hierarchy.
        """

        from os import sep, path, makedirs

        plotfile=None
        if plots_dir:
            if not path.exists(plots_dir):
                makedirs(plots_dir)
            plotfile = plots_dir + sep + title_prefix + plot['title'] + '-' + str(self.hierarchy.get_namespace()) + '.png'
        fig = self.hierarchy.hierarchy_plots(title=title_prefix + plot['title'], plot_items=plot['plot_items'], figsize=plots_figsize, file=plotfile, history=history_data)
            # import matplotlib.pyplot as plt
            # plt.close(fig)  # Close the figure here
    
        return fig
    

    



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/13_microgrid.ipynb.

# %% auto 0
__all__ = ['MicroGridEnvPlus', 'MicroGridEnv0Plus']

# %% ../nbs/13_microgrid.ipynb 4
import random
import gym.spaces as spaces
import numpy as np
import drl_microgrid_ems.tcl_env_dqn as mg0
import drl_microgrid_ems.tcl_env_dqn_1 as mg


# %% ../nbs/13_microgrid.ipynb 5
class MicroGridEnvPlus(mg.MicroGridEnv):
    def __init__(self, **kwargs):
        """
        Overriding MicroGridEnv constructor.
        """
        seed = kwargs.get("seed", None)
        if seed:
            self.seed(seed)
        pass

    def initialise(self, properties=None, **kwargs):
        # self.day0 = 1
        # self.dayn = 10

        if 'initial_seed' in properties:
            random.seed(properties['initial_seed'])

        self.day = None
        self.day_list = []
        if 'day_mode' in properties:
            self.day_mode = properties['day_mode']
        else:
            raise Exception('day_mode property for MicroGrid must be defined')
                
        # Get number of iterations and TCLs from the 
        # parameters (we have to define it through kwargs because 
        # of how Gym works...)

        self.iterations = kwargs.get("iterations", mg.DEFAULT_ITERATIONS)
        self.num_tcls = kwargs.get("num_tcls", mg.DEFAULT_NUM_TCLS)
        print(self.num_tcls)
        self.avg_tcl_power = kwargs.get("tcl_power", mg.DEFAULT_AVGTCLPOWER)
        self.tcl_sale_price = kwargs.get("tcl_price", mg.DEFAULT_TCL_SALE_PRICE)
        self.num_loads = kwargs.get("num_loads", mg.DEFAULT_NUM_LOADS)
        self.typical_load = kwargs.get("base_load", mg.DEFAULT_BASE_LOAD)
        self.market_price = kwargs.get("normal_price", mg.DEFAULT_MARKET_PRICE)
        self.temperatures = kwargs.get("temperatures", mg.DEFAULT_TEMPERATURS)
        self.price_tiers = kwargs.get("price_tiers", mg.DEFAULT_PRICE_TIERS)
        self.day0 = kwargs.get("day0", mg.DEFAULT_DAY0)
        self.dayn = kwargs.get("dayn", self.day0+1)
        self.power_cost = kwargs.get("power_cost", mg.DEFAULT_WIND_POWER_COST)
        self.down_reg = kwargs.get("down_reg", mg.DEFAULT_DOWN_REG)
        self.up_reg = kwargs.get("up_reg", mg.DEFAULT_UP_REG)
        self.imp_fees = kwargs.get("imp_fees", mg.DEFAULT_TRANSFER_PRICE_IMPORT)
        self.exp_fees = kwargs.get("exp_fees", mg.DEFAULT_TRANSFER_PRICE_EXPORT)
        self.bat_capacity = kwargs.get("battery_capacity", mg.DEFAULT_BAT_CAPACITY)
        self.max_discharge = kwargs.get("max_discharge", mg.DEFAULT_MAX_DISCHARGE)
        self.max_charge = kwargs.get("max_charge", mg.DEFAULT_MAX_CHARGE)

        if 'day0' in properties:
            self.day0 = properties['day0']
        if 'dayn' in properties:
            self.dayn = properties['dayn']


        if 'initial_day' in properties:
            self.day = properties['initial_day']
            self.set_day_list(mode=self.day_mode)
        else:
            self.day = random.randint(0,10)


        # The current timestep
        self.time_step = 0

        # The cluster of TCLs to be controlled.
        # These will be created in reset()
        self.tcls_parameters = []
        # The cluster of loads.
        # These will be created in reset()
        self.loads_parameters = []

        self.generation = mg.Generation(kwargs.get("generation_data", mg.DEFAULT_POWER_GENERATED))
        self.grid = mg.Grid(down_reg=self.down_reg,up_reg=self.up_reg, exp_fees=self.exp_fees, imp_fees=self.imp_fees)
        self.battery = mg.Battery(capacity=self.bat_capacity, useD=0.9, dissipation=0.001, rateC=0.9, maxDD=self.max_discharge, chargeE=self.max_charge)

        self.tcls = [self._create_tcl(*self._create_tcl_parameters()) for _ in range(self.num_tcls)]
        self.loads = [self._create_load(*self._create_load_parameters()) for _ in range(self.num_loads)]

        self.action_space_sep = spaces.Box(low=0, high=1, dtype=np.float32,
                                       shape=(13,))
        self.action_space = spaces.Discrete(80)

        # Observations: A vector of TCLs SoCs + loads +battery soc+ power generation + price + temperature + time of day
        self.observation_space = spaces.Box(low=-100, high=100, dtype=np.float32,
                                            shape=(self.num_tcls + 7,))


    def reset(self,day=None):
        """
        Create new TCLs, and return initial state.
        Note: Overrides previous TCLs
        """
        # if day == None:
        #     self.day= random.randint(self.day0,self.dayn)
        # else:
        #     self.day = day
        # print("Day:", self.day)

        if day==None:
            self.day = self.get_day()
        else:
            self.day = day


        self.time_step = 0

        self.high_price = 0

        return self._build_state()

    def reset_all(self,day=None):
        """
        Create new TCLs, and return initial state.
        Note: Overrides previous TCLs
        """
        # if day == None:
        #     # self.day = random.randint(self.day0, self.dayn-1)
        #     self.day= self.day0
        # else:
        #     self.day = day
        # print("Day:", self.day)

        if day==None:
            self.day = self.get_day()
        else:
            self.day = day

        self.time_step = 0
        self.battery.reset()
        self.high_price = 0
        self.tcls.clear()
        self.loads.clear()
        self.tcls = [self._create_tcl(*self._create_tcl_parameters()) for _ in range(self.num_tcls)]
        self.loads = [self._create_load(*self._create_load_parameters()) for _ in range(self.num_loads)]


        return self._build_state()



    def set_day_list(self, mode=None):
        if isinstance(mode, list):
            self.day_list =  mode.copy()
        else:
            self.day_list =  [ i for i in range(self.day0, self.dayn+1, 1)]
            if 'random' == mode:
                random.shuffle(self.day_list)

    def get_day(self):
        if len(self.day_list) == 0:
            self.set_day_list(self.day_mode)

        day = self.day_list[0]
        self.day_list.remove(day)

        return day


    def seed(self, s):
        """
        Set the random seed for consistent experiments
        """
        self.seedy(s)



# %% ../nbs/13_microgrid.ipynb 6
class MicroGridEnv0Plus(mg0.MicroGridEnv0):
    def __init__(self, **kwargs):
        """
        Overriding MicroGridEnv constructor.
        """
        seed = kwargs.get("seed", None)
        if seed:
            self.seed(seed)
        pass

    def initialise(self, properties=None, **kwargs):
        # self.day0 = 1
        # self.dayn = 10

        if 'day0' in properties:
            self.day0 = properties['day0']
        if 'dayn' in properties:
            self.dayn = properties['dayn']

        if 'initial_seed' in properties:
            random.seed(properties['initial_seed'])
        # Get number of iterations and TCLs from the 
        # parameters (we have to define it through kwargs because 
        # of how Gym works...)
        self.iterations = kwargs.get("iterations", mg0.DEFAULT_ITERATIONS)
        self.num_tcls = kwargs.get("num_tcls", mg0.DEFAULT_NUM_TCLS)
        self.num_loads = kwargs.get("num_loads", mg0.DEFAULT_NUM_LOADS)
        self.prices = kwargs.get("prices", mg0.DEFAULT_PRICES)
        self.temperatures = kwargs.get("temperatures", mg0.DEFAULT_TEMPERATURS)
        self.base_load = kwargs.get("base_load", mg0.BASE_LOAD)
        self.price_tiers = kwargs.get("price_tiers", mg0.PRICE_TIERS)

        # calculated here rather than repeatedly in _build_state
        self.temperatures_min = min(self.temperatures)
        self.temperatures_divisor = max(self.temperatures)-min(self.temperatures)

        self.day = None
        self.day_list = []
        if 'day_mode' in properties:
            self.day_mode = properties['day_mode']
        else:
            raise Exception('day_mode property for MicroGrid must be defined')
    
        if 'initial_day' in properties:
            self.day = properties['initial_day']
            self.set_day_list(mode=self.day_mode)
        else:
            self.day = random.randint(0,10)
            
        # The current timestep
        self.time_step = 0

        # The cluster of TCLs to be controlled.
        # These will be created in reset()
        self.tcls_parameters = []
        self.tcls = []
        # The cluster of loads.
        # These will be created in reset()
        self.loads_parameters = []
        self.loads = []

        self.generation = mg0.Generation(mg0.MAX_GENERATION)
        self.grid = mg0.Grid()

        # calculated here rather than repeatedly in _build_state
        self.grid_buy_prices_min = min(self.grid.buy_prices)
        self.grid_buy_prices_divisor = max(self.grid.buy_prices) - min(self.grid.buy_prices)

        for i in range(self.num_tcls):
            self.tcls_parameters.append(self._create_tcl_parameters())

        for i in range(self.num_loads):
            self.loads_parameters.append(self._create_load_parameters())

        self.action_space = spaces.Discrete(80)
        
        # Observations: A vector of TCLs SoCs + loads +battery soc+ power generation + price + temperature + time of day
        self.observation_space = spaces.Box(low=-100, high=100, dtype=np.float32, 
                    shape=(1  + 7,))


    def _build_state(self):
        """ 
        Return current state representation as one vector.
        Returns:
            state: 1D state vector, containing state-of-charges of all TCLs, Loads, current battery soc, current power generation,
                   current temperature, current price and current time (hour) of day
        """
        # SoCs of all TCLs binned + current temperature + current price + time of day (hour)
        socs = np.array([tcl.SoC for tcl in self.tcls])
        # Scaling between -1 and 1
        socs = (socs+np.ones(shape=socs.shape)*4)/(1+4)
        socs=np.average(socs)

        # loads = np.array([l.load(self.time_step) for l in self.loads])
        # loads = sum([l.load(self.time_step) for l in self.loads])
        # # Scaling loads
        # loads = (loads-(min(BASE_LOAD)+2)*DEFAULT_NUM_LOADS)/((max(BASE_LOAD)+4-min(BASE_LOAD)-2)*DEFAULT_NUM_LOADS)

        loads = mg0.BASE_LOAD[(self.time_step) % 24]
        loads = (loads - min(mg0.BASE_LOAD)) / (max(mg0.BASE_LOAD) - min(mg0.BASE_LOAD))


        current_generation = self.generation.current_generation(self.day+self.time_step)
        current_generation /= self.generation.max_capacity

        temperature = self.temperatures[self.day+self.time_step]
        temperature = (temperature-self.temperatures_min)/self.temperatures_divisor

        price = self.grid.buy_prices[self.day+self.time_step]
        price = (price - self.grid_buy_prices_min) / self.grid_buy_prices_divisor

        high_price = self.high_price/(4 * self.iterations)

        time_step = (self.time_step)/24

        state = np.array([socs, loads, high_price, self.battery.SoC, current_generation,
                         temperature,
                         price,
                         time_step])
        return state


    def reset(self,day=None):
        """
        Create new TCLs, and return initial state.
        Note: Overrides previous TCLs
        """
        if day==None:
            self.day = self.get_day()
        else:
            self.day = day

        # print("Day:",self.day)
        self.time_step = 0
        self.battery = self._create_battery()
        self.energy_sold = 0
        self.energy_bought = 0
        self.energy_generated = 0
        self.control=0
        self.sale_price = mg0.PRICE_TIERS[2]
        self.high_price = 0
        self.tcls.clear()
        # initial_tcl_temperature = random.normalvariate(12, 5)
        initial_tcl_temperature = 12

        for i in range(self.num_tcls):
            parameters = self.tcls_parameters[i]
            self.tcls.append(self._create_tcl(parameters[0],parameters[1],parameters[2],parameters[3],initial_tcl_temperature))

        self.loads.clear()
        for i in range(self.num_loads):
            parameters = self.loads_parameters[i]
            self.loads.append(self._create_load(parameters[0],parameters[1]))

        self.battery = self._create_battery()
        return self._build_state()



    def set_day_list(self, mode=None):
        if isinstance(mode, list):
            self.day_list =  mode.copy()
        else:
            self.day_list =  [ i for i in range(self.day0, self.dayn+1, 1)]
            if 'random' == mode:
                random.shuffle(self.day_list)

    def get_day(self):
        if len(self.day_list) == 0:
            self.set_day_list(self.day_mode)

        day = self.day_list[0]
        self.day_list.remove(day)

        return day

    



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/03_nodes.ipynb.

# %% auto 0
__all__ = ['ControlUnitIndices', 'PCTNode', 'PCTNodeData']

# %% ../nbs/03_nodes.ipynb 4
import enum
import uuid
from .putils import UniqueNamer
from .putils import FunctionsList
from .putils import dynamic_module_import
from .functions import FunctionFactory
from .functions import HPCTFUNCTION

import json
from .functions import *

# %% ../nbs/03_nodes.ipynb 5
class ControlUnitIndices(enum.IntEnum):
   PER_INDEX = 0
   OUT_INDEX = 1
   REF_INDEX = 2
   ACT_INDEX = 3

# %% ../nbs/03_nodes.ipynb 6
class PCTNode():
    "A single PCT controller."
    def __init__(self, reference=None, perception=None, comparator=None, output=None, default=True, 
                 name="pctnode", history=False, build_links=False, mode=0, namespace=None, **pargs):
        # mode
        # 0 - per:var, ref:con, com:sub, out:prop
        # 1 - per:ws, ref:ws, com:sub, out:prop
        # 2 - per:ws, ref:con, com:sub, out:prop
        # 3 - per:ws, ref:ws, com:sub, out:ws
        # 4 - per:ws, ref:con, com:sub, out:ws
        # 5 - per:ws, ref:con, com:sub, out:smws
        # 6 - per:ws, ref:ws, com:sub, out:smws
        
        if namespace ==None:
            namespace = uuid.uuid1()
        self.namespace=namespace
        
        self.links_built = False
        self.history = None
        if history:
            self.history = PCTNodeData()
        self.name = UniqueNamer.getInstance().get_name(self.namespace, name)
        FunctionsList.getInstance().add_function(self.namespace, self)
        if default:
            if perception==None:
                perception = FunctionFactory.createFunctionWithNamespace(PCTNode.get_function_type(mode, HPCTFUNCTION.PERCEPTION), namespace=namespace)
            self.perceptionCollection = [perception]

            if reference==None:
                reference = FunctionFactory.createFunctionWithNamespace(PCTNode.get_function_type(mode, HPCTFUNCTION.REFERENCE), namespace=namespace)
            self.referenceCollection = [reference]

            if comparator==None:
                comparator = FunctionFactory.createFunctionWithNamespace(PCTNode.get_function_type(mode, HPCTFUNCTION.COMPARATOR), namespace=namespace)
            self.comparatorCollection = [comparator]

            if output==None:
                output = FunctionFactory.createFunctionWithNamespace(PCTNode.get_function_type(mode, HPCTFUNCTION.OUTPUT), namespace=namespace)
            self.outputCollection = [output]
            
            if build_links:
                self.build_links()
        else:
            self.perceptionCollection = None
            self.referenceCollection = None
            self.comparatorCollection = None
            self.outputCollection = None

    
    def __call__(self, verbose=False):
        if not self.links_built:
            self.build_links()
            
        if self.referenceCollection:
            for referenceFunction in self.referenceCollection:
                referenceFunction(verbose)               

        if self.perceptionCollection:
            for perceptionFunction in self.perceptionCollection:
                perceptionFunction(verbose)
                    
        if self.comparatorCollection:
            for comparatorFunction in self.comparatorCollection:
                comparatorFunction(verbose)

        if self.outputCollection:
            for outputFunction in self.outputCollection:
                outputFunction(verbose)                
            self.output = self.outputCollection[-1].get_value()
        else:
            self.output = 0
        
        if verbose:
            print()
            
        if not self.history == None:
            self.history.add_data(self)
            
        return self.output
    
    def __str__(self):
        return str(self.__dict__)
    
    def is_reference_empty(self):
        if self.referenceCollection:
            return False

        return True

    def is_empty(self):
        if self.perceptionCollection:
            return False
        if self.referenceCollection:
            return False
        if self.comparatorCollection:
            return False
        if self.outputCollection:
            return False

        return True
    
    def get_summary(self):
        
        for referenceFunction in self.referenceCollection:
            r= referenceFunction.output_string()               

        for perceptionFunction in self.perceptionCollection:
            p = perceptionFunction.output_string()
                    
        for comparatorFunction in self.comparatorCollection:
            c = comparatorFunction.output_string()

        for outputFunction in self.outputCollection:
            o = outputFunction.output_string()
        
        return " ".join((r, p, c, o))
    
    def set_name(self, name):
        self.name=name
    
    def get_name(self):
        return self.name

    def get_function_from_collection(self, collection, position=-1):
        if collection == HPCTFUNCTION.REFERENCE:
            func = self.referenceCollection[position]
            return func 

        if collection == HPCTFUNCTION.PERCEPTION:
            func = self.perceptionCollection[position]
            return func 
 
        if collection == HPCTFUNCTION.COMPARATOR:
            func = self.comparatorCollection[position]
            return func 
        
        if collection == HPCTFUNCTION.OUTPUT:
            func = self.outputCollection[position]
            return func 
        
        return func 
    
    def reset(self):
        if self.referenceCollection:
            for func in self.referenceCollection:
                func.reset_value()               

        if self.perceptionCollection:
            for func in self.perceptionCollection:
                func.reset_value()               

        if self.comparatorCollection:
            for func in self.comparatorCollection:
                func.reset_value()              

        if self.outputCollection:
            for func in self.outputCollection:
                func.reset_value()               
    

    def delete_perception(self):
        for func in self.perceptionCollection:
            FunctionsList.getInstance().delete_function(func.namespace, func.name)

        for ctr in range(len(self.perceptionCollection)-1, -1, -1):
            del self.perceptionCollection[ctr]

        self.perceptionCollection = None


    def delete_output(self):
        for func in self.outputCollection:
            FunctionsList.getInstance().delete_function(func.namespace, func.name)

        for ctr in range(len(self.outputCollection)-1, -1, -1):
            del self.outputCollection[ctr]

        self.outputCollection = None


    def delete_comparator(self):
        for func in self.comparatorCollection:
            FunctionsList.getInstance().delete_function(func.namespace, func.name)

        for ctr in range(len(self.comparatorCollection)-1, -1, -1):
            del self.comparatorCollection[ctr]

        self.comparatorCollection = None

    def delete_reference(self):
        for func in self.referenceCollection:
            FunctionsList.getInstance().delete_function(func.namespace, func.name)

        for ctr in range(len(self.referenceCollection)-1, -1, -1):
            del self.referenceCollection[ctr]

        self.referenceCollection = None

    def delete_contents(self):
        for func in self.referenceCollection:
            FunctionsList.getInstance().delete_function(func.namespace, func.name)

        for func in self.perceptionCollection:
            FunctionsList.getInstance().delete_function(func.namespace, func.name)
                    
        for func in self.comparatorCollection:
            FunctionsList.getInstance().delete_function(func.namespace, func.name)

        for func in self.outputCollection:
            FunctionsList.getInstance().delete_function(func.namespace, func.name)
            
    def reset_checklinks(self, val=True):
        for func in self.referenceCollection:
            func.reset_checklinks(val)               

        for func in self.perceptionCollection:
            func.reset_checklinks(val)               
                    
        for func in self.comparatorCollection:
            func.reset_checklinks(val)               

        for func in self.outputCollection:
            func.reset_checklinks(val)               
                    

    def get_function(self, collection, position=-1):
        if collection == "reference":
            func = self.referenceCollection[position]

        if collection == "perception":
            func = self.perceptionCollection[position]
 
        if collection == "comparator":
            func = self.comparatorCollection[position]
        
        if collection == "output":
            func = self.outputCollection[position]
        
        return func 

    def get_perception_value(self, position=-1):
        return self.perceptionCollection[position].get_value()
    
    def set_perception_value(self, value, position=-1):
        self.perceptionCollection[position].set_value(value)
    
    def add_link(self, collection, link):
        if collection == "reference":
            self.referenceCollection[0].add_link(link)

        if collection == "perception":
            self.perceptionCollection[0].add_link(link)
 
        if collection == "comparator":
            self.comparatorCollection[-1].add_link(link)
        
        if collection == "output":
            self.outputCollection[-1].add_link(link)
            
    # build links between each other in a collection, between com and ref/per and from out to com
    def build_links(self):
        if self.is_empty():       
            self.links_built = True
            return

        if len(self.referenceCollection)>0:
            link = self.referenceCollection[0]
            for i in range (1, len(self.referenceCollection)):
                self.referenceCollection[i].add_link(link)               
                link = self.referenceCollection[i]

        if len(self.perceptionCollection)>0:
            link = self.perceptionCollection[0]
            for i in range (1, len(self.perceptionCollection)):
                self.perceptionCollection[i].add_link(link)               
                link = self.perceptionCollection[i]

        self.comparatorCollection[0].add_link(self.referenceCollection[-1])
        self.comparatorCollection[0].add_link(self.perceptionCollection[-1])

        if len(self.comparatorCollection)>1:
            link = self.comparatorCollection[1]
            for i in range (1, len(self.comparatorCollection)):
                self.comparatorCollection[i].add_link(link)               
                link = self.comparatorCollection[i]

        self.outputCollection[0].add_link(self.comparatorCollection[-1])

        if len(self.outputCollection)>0:
            link = self.outputCollection[0]
            for i in range (1, len(self.outputCollection)):
                self.outputCollection[i].add_link(link)               
                link = self.outputCollection[i]

        self.links_built = True

    def run(self, steps=None, verbose=False):
        for i in range(steps):
            out = self(verbose)
        return out
    
    def set_output(self, value):
        self.outputCollection[-1].set_value(value)
        
    def get_output_function(self):
        return self.outputCollection[-1]

    def has_reference_function(self):
        if self.referenceCollection is None:
            return False
        return True

    def get_reference_function(self):
        return self.referenceCollection[-1]

    def get_perception_function(self):
        return self.perceptionCollection[-1]

    def set_reference(self, value):
        self.referenceCollection[-1].set_value(value)

    
    def set_function_name(self, collection, name, position=-1):
        if collection == "reference":
            self.referenceCollection[position].set_name(name)

        if collection == "perception":
            self.perceptionCollection[position].set_name(name)
 
        if collection == "comparator":
            self.comparatorCollection[position].set_name(name)
        
        if collection == "output":
            self.outputCollection[position].set_name(name)
            

    def replace_function(self, collection, function, position=-1):
        if collection == "reference":
            self.referenceCollection[position] = function

        if collection == "perception":
            self.perceptionCollection[position]  = function
 
        if collection == "comparator":
            self.comparatorCollection[position] = function
        
        if collection == "output":
            self.outputCollection[position] = function
            
        

    def insert_function(self, collection, function, position=-1):
        if collection == "reference":
            self.referenceCollection[position] = function

        if collection == "perception":
            self.perceptionCollection[position]  = function
 
        if collection == "comparator":
            self.comparatorCollection[position] = function
        
        if collection == "output":
            self.outputCollection[position] = function


    def check_namespace(self, higher_namespace=None):

        if higher_namespace:
            if higher_namespace != self.namespace:
                print(self.name, type(self).__name__)
                print(f'This namespace {self.namespace} does not equal higher namespace {higher_namespace}')
                raise Exception(f'This namespace {self.namespace} does not equal higher namespace {higher_namespace}')

        for referenceFunction in self.referenceCollection:
            referenceFunction.check_namespace(higher_namespace=higher_namespace)   
        
        for perceptionFunction in self.perceptionCollection:
            perceptionFunction.check_namespace(higher_namespace=higher_namespace)
        
        for comparatorFunction in self.comparatorCollection:
            comparatorFunction.check_namespace(higher_namespace=higher_namespace)
        
        for outputFunction in self.outputCollection:
            outputFunction.check_namespace(higher_namespace=higher_namespace)

            
    def summary(self, build=True, extra=False, higher_namespace=None):
        if build:
            if not self.links_built:
                self.build_links()

        print(self.name, type(self).__name__, self.namespace)
        print("----------------------------")

        if higher_namespace:
            if higher_namespace != self.namespace:
                raise Exception(f'This namespace {self.namespace} does not equal higher namespace {higher_namespace}')

        print("REF:", end=" ")
        if self.referenceCollection:
            for referenceFunction in self.referenceCollection:
                referenceFunction.summary(extra=extra, higher_namespace=higher_namespace)   
        
        print("PER:", end=" ")
        for perceptionFunction in self.perceptionCollection:
            perceptionFunction.summary(extra=extra, higher_namespace=higher_namespace)
        
        print("COM:", end=" ")
        if self.comparatorCollection:     
            for comparatorFunction in self.comparatorCollection:
                comparatorFunction.summary(extra=extra, higher_namespace=higher_namespace)      
        
        print("OUT:", end=" ")
        if self.outputCollection:
            for outputFunction in self.outputCollection:
                outputFunction.summary(extra=extra, higher_namespace=higher_namespace)
        
        print("----------------------------")
        
        
    def graph(self, layer=0, layout={'r':2,'c':1,'p':2, 'o':0}):
        import networkx as nx

        graph = nx.DiGraph()
        
        self.set_graph_data(graph, layer=layer, layout=layout)
                
        return graph
    

    def validate(self, num_lower_perceptions=None, num_higher_outputs=None):
        
        if num_higher_outputs is not None:
            for func in self.referenceCollection:
                func.validate(num_higher_outputs)
                
        for func in self.perceptionCollection:
            func.validate(num_lower_perceptions)
    
    def remove_links(self):
        for referenceFunction in self.referenceCollection:
            referenceFunction.remove_links()  
               
        for perceptionFunction in self.perceptionCollection:
            perceptionFunction.remove_links()
        
        # for outputFunction in self.outputCollection:
        #     outputFunction.remove_links()


    def list_link_names(self, link_names):
        if self.referenceCollection:
            for func in self.referenceCollection:
                for link in func.links:
                    if isinstance(link, str):
                        link_names.append(link)
                    else:
                        link_names.append(link.get_name())
        if self.comparatorCollection:
            for func in self.comparatorCollection:
                for link in func.links:
                    if isinstance(link, str):
                        link_names.append(link)
                    else:
                        link_names.append(link.get_name())
        if self.perceptionCollection:    
            for func in self.perceptionCollection:
                for link in func.links:
                    if isinstance(link, str):
                        link_names.append(link)
                    else:
                        link_names.append(link.get_name())
        if self.outputCollection:    
            for func in self.outputCollection:
                for link in func.links:
                    if isinstance(link, str):
                        link_names.append(link)
                    else:
                        link_names.append(link.get_name())


    def consolidate(self, link_names):
        if self.outputCollection:    
            outputFunction = self.outputCollection[-1]
            name = outputFunction.get_name()
            if name not in link_names:
                self.delete_output()
                self.delete_comparator()
                self.delete_reference()
                # print(self.outputCollection)
                # print(self.comparatorCollection)
                # print(self.referenceCollection)

        perceptionFunction = self.perceptionCollection[-1]
        name = perceptionFunction.get_name()
        if name not in link_names:
            self.delete_perception()


        


    def clear_values(self):
        for referenceFunction in self.referenceCollection:
            referenceFunction.value = 0  
        
        for comparatorFunction in self.comparatorCollection:
            comparatorFunction.value = 0
        
        for perceptionFunction in self.perceptionCollection:
            perceptionFunction.value = 0
        
        for outputFunction in self.outputCollection:
            outputFunction.value  = 0
    
    def change_namespace(self, namespace):
        self.namespace=namespace
        
        name = self.name
        newname = UniqueNamer.getInstance().get_name(namespace, self.name)
        if name != newname:
            raise Exception(f'Names should not be different {name} {newname}')
        FunctionsList.getInstance().add_function(namespace, self)

        for referenceFunction in self.referenceCollection:
            referenceFunction.change_namespace(namespace)
        
        for comparatorFunction in self.comparatorCollection:
            comparatorFunction.change_namespace(namespace)
        
        for perceptionFunction in self.perceptionCollection:
            perceptionFunction.change_namespace(namespace)
        
        for outputFunction in self.outputCollection:
            outputFunction.change_namespace(namespace)
        
 
#     def set_graph_data_node(self, graph, layer=0):
#         graph.add_node(self.name, layer=layer)

#         for referenceFunction in self.referenceCollection:
#             referenceFunction.set_graph_data(graph, layer+2)   

#         for perceptionFunction in self.perceptionCollection:
#             perceptionFunction.set_graph_data(graph, layer+2)

            
    def set_graph_data(self, graph, layer=0, layout={'r':2,'c':1,'p':2, 'o':0}):
        if self.referenceCollection:
            for referenceFunction in self.referenceCollection:
                referenceFunction.set_graph_data(graph, layer+layout['r'])   
        if self.comparatorCollection:
            for comparatorFunction in self.comparatorCollection:
                comparatorFunction.set_graph_data(graph, layer+layout['c'])
        if self.perceptionCollection:
            for perceptionFunction in self.perceptionCollection:
                perceptionFunction.set_graph_data(graph, layer+layout['p'])
        if self.outputCollection:
            for outputFunction in self.outputCollection:
                outputFunction.set_graph_data(graph, layer+layout['o'])

            
    def set_graph_data_funcdata(self, graph, layer=0, layout={'r':2,'c':1,'p':2, 'o':0}):
        if self.referenceCollection:       
            for referenceFunction in self.referenceCollection:
                referenceFunction.set_graph_data_funcdata(graph, layer+layout['r'])   
        
        if self.comparatorCollection:
            for comparatorFunction in self.comparatorCollection:
                comparatorFunction.set_graph_data_funcdata(graph, layer+layout['c'])
        
        if self.perceptionCollection:
            for perceptionFunction in self.perceptionCollection:
                perceptionFunction.set_graph_data_funcdata(graph, layer+layout['p'])
        
        if self.outputCollection:
            for outputFunction in self.outputCollection:
                outputFunction.set_graph_data_funcdata(graph, layer+layout['o'])
            
            
    def get_edge_labels(self, labels):
        if self.referenceCollection:
            for func in self.referenceCollection:
                func.get_weights_labels(labels)
        if self.comparatorCollection:
            for func in self.comparatorCollection:
                func.get_weights_labels(labels)
        if self.perceptionCollection:                    
            for func in self.perceptionCollection:
                func.get_weights_labels(labels)
        if self.outputCollection:                    
            for func in self.outputCollection:
                func.get_weights_labels(labels)

    def get_edge_labels_funcdata(self, labels):
        if self.referenceCollection:
            for func in self.referenceCollection:
                func.get_weights_labels_funcdata(labels)
        if self.comparatorCollection:
            for func in self.comparatorCollection:
                func.get_weights_labels_funcdata(labels)                    
        if self.perceptionCollection:                    
            for func in self.perceptionCollection:
                func.get_weights_labels_funcdata(labels)                    
        if self.outputCollection:                    
            for func in self.outputCollection:
                func.get_weights_labels_funcdata(labels)
            
            
    def get_node_list(self, node_list):
    
        for func in self.referenceCollection:
            node_list[func.get_name()] = self.name

        for func in self.comparatorCollection:
            node_list[func.get_name()] = self.name

        for func in self.perceptionCollection:
            node_list[func.get_name()] = self.name

        for func in self.outputCollection:
            node_list[func.get_name()] = self.name

    def get_parameters_list(self):

        ref_list = []
        if self.referenceCollection:
            for func in self.referenceCollection:
                ref_list.append(func.get_parameters_list())

        per_list = []
        if self.perceptionCollection:
            for func in self.perceptionCollection:
                per_list.append(func.get_parameters_list())

        out_list = []
        if self.outputCollection:
            for func in self.outputCollection:
                out_list.append(func.get_parameters_list())


        node_list = [ref_list,per_list,out_list]
            
        return node_list
        
    def change_link_name(self, old_name, new_name):

        for func in self.referenceCollection:
            func.links = [new_name if i==old_name else i for i in func.links ]

        for func in self.comparatorCollection:
            func.links = [new_name if i==old_name else i for i in func.links ]

        for func in self.perceptionCollection:
            func.links = [new_name if i==old_name else i for i in func.links ]

        for func in self.outputCollection:
            func.links = [new_name if i==old_name else i for i in func.links ]
        
            
        
    def set_graph_data_node(self, graph, layer=0):
        graph.add_node(self.name, layer=layer)


    def graph_node(self, layer=0):
        graph = nx.DiGraph()

        self.set_graph_data_node(graph, layer=layer)

        return graph


    def draw_node(self, with_labels=True,  font_size=12, font_weight='bold', node_color='red',  
         node_size=500, arrowsize=25, align='horizontal', file=None, figsize=(5,5), move={}):

        graph = self.graph_node()
        pos = nx.multipartite_layout(graph, subset_key="layer", align=align)        
        plt.figure(figsize=figsize)
        nx.draw(graph, pos=pos, with_labels=with_labels, font_size=font_size, font_weight=font_weight, 
                node_color=node_color,  node_size=node_size, arrowsize=arrowsize)
        
        
        
    def draw(self, with_labels=True,  font_size=12, font_weight='bold', node_color='red',  
             node_size=500, arrowsize=25, align='horizontal', file=None, figsize=(5,5), move={}):
        import networkx as nx        
        import matplotlib.pyplot as plt
        graph = self.graph()
        pos = nx.multipartite_layout(graph, subset_key="layer", align=align)        
        plt.figure(figsize=figsize)
        nx.draw(graph, pos=pos, with_labels=with_labels, font_size=font_size, font_weight=font_weight, 
                node_color=node_color,  node_size=node_size, arrowsize=arrowsize)
        
    def get_config(self, zero=1):
        config = {"type": type(self).__name__,
                    "name": self.name}        

        if self.referenceCollection:
            coll_name = 'refcoll'
            collection = self.referenceCollection
            config[coll_name] = self.get_collection_config(coll_name, collection, zero)       
        if self.perceptionCollection:
            coll_name = 'percoll'
            collection = self.perceptionCollection
            config[coll_name] = self.get_collection_config(coll_name, collection, zero)       
        if self.comparatorCollection:
            coll_name = 'comcoll'
            collection = self.comparatorCollection
            config[coll_name] = self.get_collection_config(coll_name, collection, zero)       
        if self.outputCollection:
            coll_name = 'outcoll'
            collection = self.outputCollection
            config[coll_name] = self.get_collection_config(coll_name, collection, zero)       
        
        return config       

    def get_collection_config(self, coll_name, collection, zero=1):
        coll = {}
        ctr=0
        for func in collection:
            coll[str(ctr)] = func.get_config(zero=zero)
            ctr+=1            
        return coll
    
    def save(self, file=None, indent=4):
        jsondict = json.dumps(self.get_config(), indent=indent)
        f = open(file, "w")
        f.write(jsondict)
        f.close()
        
    @classmethod
    def load(cls, file):      
        with open(file) as f:
            config = json.load(f)
        return cls.from_config(config, reference=True, comparator=True, perception=True, output=True)

    @classmethod
    def from_config(cls, config=None, namespace=None, node=None, reference=False, comparator=False, perception=False, output=False, history=False):
        "Create a node from JSON dictionary configuration."
        if node is None:
            node = PCTNode(default=False, name=config['name'], namespace=namespace, history=history)

        namespace= node.namespace

        if reference and 'refcoll' in config:
            node.referenceCollection = []
            collection = node.referenceCollection
            coll_dict = config['refcoll']
            PCTNode.collection_from_config(collection, coll_dict, namespace)

        if perception and 'percoll' in config:
            node.perceptionCollection = []
            collection = node.perceptionCollection
            coll_dict = config['percoll']
            PCTNode.collection_from_config(collection, coll_dict, namespace)

        if comparator and 'comcoll' in config:
            node.comparatorCollection = []
            collection = node.comparatorCollection
            coll_dict = config['comcoll']
            PCTNode.collection_from_config(collection, coll_dict, namespace)

        if output and 'outcoll' in config:
            node.outputCollection = []
            collection = node.outputCollection
            coll_dict = config['outcoll']
            PCTNode.collection_from_config(collection, coll_dict, namespace)

        node.links_built = True
        return node
    
#     @classmethod
#     def from_config(cls, config, namespace=None, history=False):
            
#         node = PCTNode(default=False, name=config['name'], namespace=namespace, history=history)
#         namespace= node.namespace 
#         node.referenceCollection = []        
#         collection = node.referenceCollection
#         coll_dict = config['refcoll']
#         PCTNode.collection_from_config(collection, coll_dict, namespace)

#         node.perceptionCollection = []
#         collection = node.perceptionCollection
#         coll_dict = config['percoll']
#         PCTNode.collection_from_config(collection, coll_dict, namespace)
        
#         node.comparatorCollection = []
#         collection = node.comparatorCollection
#         coll_dict = config['comcoll']
#         PCTNode.collection_from_config(collection, coll_dict, namespace)

#         node.outputCollection = []
#         collection = node.outputCollection
#         coll_dict = config['outcoll']
#         PCTNode.collection_from_config(collection, coll_dict, namespace)
        
#         node.links_built = True
#         return node
    
    @classmethod
    def collection_from_config(node, collection, coll_dict, namespace):
        #print("collection_from_config", coll_dict)
        for fndict_label in coll_dict:
            #print("fndict_label",fndict_label)
            
            fndict = coll_dict[fndict_label]
            #print(fndict)
            fnname = fndict.pop('type')
            if fnname.startswith("EA"):
                fnname = fnname.replace("EA", "")
            #print(fndict)
            #func = eval(fnname).from_config(fndict, namespace)
            func = FunctionFactory.createFunctionFromConfig(fnname, namespace, fndict)
            
            collection.append(func)
            
    @classmethod
    def get_function_type(cls, mode, function):
        type = 'WeightedSum'
       
        if function == HPCTFUNCTION.PERCEPTION:
            if mode == 0 :
                type = 'Variable'
                
        if function == HPCTFUNCTION.REFERENCE:
            if mode == 0 or mode == 2  or mode == 4  or mode == 5 :
                type = 'Constant'
            
        if function == HPCTFUNCTION.COMPARATOR:
            type = 'Subtract'
            
        if function == HPCTFUNCTION.OUTPUT:
            if mode < 3: 
                type = 'Proportional'
            elif mode == 5  or mode == 6 :
                type = 'SmoothWeightedSum'
                
        return type    

# %% ../nbs/03_nodes.ipynb 8
class PCTNodeData():
    "Data collected for a PCTNode"
    def __init__(self, name="pctnodedata"):
        self.data = {
            "refcoll":{}, 
            "percoll":{},
            "comcoll":{}, 
            "outcoll":{}}
        
    
    def add_data(self, node):
        if node.is_empty():
            return
        if node.referenceCollection:
            self.add_collection( node.referenceCollection, "refcoll")
        if node.perceptionCollection:
            self.add_collection( node.perceptionCollection, "percoll")
        if node.comparatorCollection:
            self.add_collection( node.comparatorCollection, "comcoll")
        if node.outputCollection:
            self.add_collection( node.outputCollection, "outcoll")

    def add_collection(self, collection, collname):
        for func in collection:            
            if self.data[collname].get(func.get_name()) == None:
                dlist=[]
                cdict={func.get_name():dlist}
                self.data[collname]=cdict
            else:
                dlist = self.data[collname][func.get_name()]
                
            dlist.append(func.get_value())



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/16_environment_processing.ipynb.

# %% auto 0
__all__ = ['wind_turbine_results', 'EnvironmentProcessingFactory', 'BaseEnvironmentProcessing',
           'WindTurbineEnvironmentProcessing', 'DummyEnvironmentProcessing', 'GenericEnvironmentProcessing',
           'GenericGymEnvironmentProcessing', 'ARCEnvironmentProcessing']

# %% ../nbs/16_environment_processing.ipynb 3
from abc import ABC, abstractmethod
from os import sep, makedirs, getenv
import time

# Ignore deprecation warnings
import warnings
warnings.filterwarnings("ignore", category=DeprecationWarning)

# %% ../nbs/16_environment_processing.ipynb 4
# from pct.yaw_module import get_comparaison_metrics, test_trad_control, test_hpct_wind, get_properties, get_indexes
from .putils import printtime, NumberStats, PCTRunProperties
from .helpers import DataManagerSingleton
from .plot import PlotARCArrays
from .hierarchy import PCTHierarchy

# %% ../nbs/16_environment_processing.ipynb 6
def wind_turbine_results(environment_properties=None, experiment=None, root=None, wt='WindTurbine', verbose=None, early=None, min=None,
                         comparisons=False, comparisons_print_plots=False, property_dir=None, property_file=None, plots=None, log_testing_to_experiment=False, hierarchy=None):

    from pct.yaw_module import get_comparaison_metrics, test_trad_control, test_hpct_wind, get_properties, get_indexes
    prefix = property_file[:property_file.find(".properties")]
    filename=wt+sep+property_dir+sep+property_file
    file = root + 'data'+sep+'ga'+sep+ filename

    wind_timeseries,start, stop, model_params,yaw_params,keep_history, rt = get_properties(environment_properties)

    history=True
    series = environment_properties['series']
    outdir='c:'+sep+'tmp'+sep+'WindTurbine'+ sep + environment_properties['range'] + sep + series + sep + prefix +sep
    makedirs(outdir, exist_ok=True)

    # lastsepIndex = filename.rfind(sep)
    # propIndex = filename.rfind('.properties')
    # filenamePrefix = filename[lastsepIndex+1:propIndex]
    # draw_file = outdir + 'draw-'+filenamePrefix+'.png'
    draw_file = False
    model_file = outdir + 'res_model.html'

    if experiment:
        from comet_ml import Artifact
        artifact = Artifact(property_file, "Properties file")
        artifact.add(file)
        experiment.log_artifact(artifact)

    (res_model, nac_pos_model, power_improvement, power_control, power_simu) = test_hpct_wind(
        file=file,plots=plots,history=history,verbose=verbose,outdir=outdir,early=early,draw_file=draw_file, model_file=model_file,
        environment_properties=environment_properties,
        start_index=model_params['start_index_test'],
        stop_index=model_params['stop_index_test'],
        experiment=experiment,
        datatype='test',
        log_testing_to_experiment=log_testing_to_experiment, min=min, hierarchy=hierarchy
        )

    if comparisons:
        start, stop = get_indexes(model_params, environment_properties)

        (res_baseline_simu, nac_pos_baseline_simu, wind_dir) = test_trad_control(
            model_params['wind_timeseries'],
            model_params['wind_timeseries_not_agg'],
            yaw_params['cycle_period'],
            start, #model_params['start_index_test'],
            stop, #model_params['stop_index_test'],
            experiment=experiment,
            datatype='baseline_simu',
            outdir=outdir
            )

        (res_baseline_logs, nac_pos_baseline_logs, wind_dir) = test_trad_control(
            model_params['wind_timeseries'],
            model_params['wind_timeseries_not_agg'],
            yaw_params['cycle_period'],
            start, #model_params['start_index_test'],
            stop, #model_params['stop_index_test'],
            experiment=experiment,
            datatype='baseline_logs',
            outdir=outdir
            )

    rel_net_prod_change=[]
    net_prod_change=[]
    power_prod_change, conso_yaw_change, net_prod_change, rel_net_prod_change,yaw_error_rel_change = None, None, None, None, None
    if comparisons :
        power_prod_change, conso_yaw_change, net_prod_change, rel_net_prod_change,yaw_error_rel_change = get_comparaison_metrics(wind_dir,power_control,power_simu,nac_pos_model, nac_pos_baseline_simu, yaw_params['yaw_rate_max'], yaw_params['yaw_consumption'], 50)    

        if comparisons_print_plots:
            fig, axs = plt.subplots(6, sharex=True, figsize=(15,25), gridspec_kw={'height_ratios': [3, 1, 1,1,1,1]})
            plt.xlabel('time (sec)', fontsize=20)
            plt.xlim(0,(model_params['stop_index_test']- model_params['start_index_test'])*yaw_params['cycle_period'])

            axs[0].plot(range(0,(model_params['stop_index_test']- model_params['start_index_test'])*yaw_params['cycle_period'] ,yaw_params['cycle_period']),wind_dir,label='wind direction')
            axs[0].plot(range(0,(model_params['stop_index_test']- model_params['start_index_test'])*yaw_params['cycle_period'] ,yaw_params['cycle_period']),nac_pos_model,linewidth=4.0, label='nacelle position RL algorithm')
            axs[0].plot(range(0,(model_params['stop_index_test']- model_params['start_index_test'])*yaw_params['cycle_period'],yaw_params['cycle_period'] ),nac_pos_baseline_simu,linewidth=4.0, label='nacelle position simulated baseline algorithm')
            axs[0].plot(range(0,(model_params['stop_index_test']- model_params['start_index_test'])*yaw_params['cycle_period'] ,yaw_params['cycle_period']),nac_pos_baseline_logs,linewidth=4.0, label='nacelle position baseline algorithm from logs')
            axs[0].legend(fontsize=20)
            axs[0].tick_params(labelsize=20)
            axs[0].set_ylabel(ylabel='angle (deg)', fontsize=20)

            axs[1].bar(range(0,(model_params['stop_index_test']- model_params['start_index_test'])*yaw_params['cycle_period'] ,yaw_params['cycle_period']*50),yaw_error_rel_change,500,linewidth=4.0, label='RL alg', align='edge')
            axs[1].get_yaxis().set_major_formatter(FuncFormatter(lambda y, _: "{:.1f}%".format(y)))
            axs[1].legend(fontsize=20)
            axs[1].tick_params(labelsize=20)
            axs[1].set_ylabel(ylabel='yaw misaligment \ndecrease (per cent)', fontsize=20,)

            axs[2].bar(range(0,(model_params['stop_index_test']- model_params['start_index_test'])*yaw_params['cycle_period'] ,yaw_params['cycle_period']*50),power_prod_change,500,linewidth=4.0, label='RL alg', align='edge')
            plt.gca().yaxis.set_major_formatter(FuncFormatter(lambda y, _: "{:.1f}%".format(y)))
            axs[2].legend(fontsize=20)
            axs[2].tick_params(labelsize=20)
            axs[2].set_ylabel(ylabel='power output \nincrease (kW)', fontsize=20,)

            axs[3].bar(range(0,(model_params['stop_index_test']- model_params['start_index_test'])*yaw_params['cycle_period'] ,yaw_params['cycle_period']*50),conso_yaw_change,500,linewidth=4.0, label='RL alg', align='edge')
            plt.gca().yaxis.set_major_formatter(FuncFormatter(lambda y, _: "{:.1f}%".format(y)))
            axs[3].legend(fontsize=20)
            axs[3].tick_params(labelsize=20)
            axs[3].set_ylabel(ylabel='yaw consumption \nincrease (kW)', fontsize=20,)

            axs[4].bar(range(0,(model_params['stop_index_test']- model_params['start_index_test'])*yaw_params['cycle_period'] ,yaw_params['cycle_period']*50),net_prod_change,500,linewidth=4.0, label='RL alg', align='edge')
            axs[4].legend(fontsize=20)
            axs[4].tick_params(labelsize=20)
            axs[4].set_ylabel(ylabel='net power output \nincrease (kW)', fontsize=20,)

            axs[5].bar(range(0,(model_params['stop_index_test']- model_params['start_index_test'])*yaw_params['cycle_period'] ,yaw_params['cycle_period']*50),rel_net_prod_change,500,linewidth=4.0, label='RL alg', align='edge')
            axs[5].get_yaxis().set_major_formatter(FuncFormatter(lambda y, _: "{:.1f}%".format(y)))
            axs[5].legend(fontsize=20)
            axs[5].tick_params(labelsize=20)
            axs[5].set_ylabel(ylabel='net power output \nincrease (per cent)', fontsize=20,)

            fig.tight_layout()
            plt.savefig(outdir + f'{series}_dataset',dpi=600)


            fig, axs = plt.subplots(2, sharex=True)
            axs[0].plot(range(model_params['start_index']*yaw_params['cycle_period'],model_params['stop_index']*yaw_params['cycle_period'], yaw_params['cycle_period'] ),wind_timeseries['wind_direction'][model_params['start_index']:model_params['stop_index']],label='train')
            axs[0].plot(range(model_params['start_index_test']*yaw_params['cycle_period'],model_params['stop_index_test']*yaw_params['cycle_period'],yaw_params['cycle_period'] ),wind_timeseries['wind_direction'][model_params['start_index_test']: model_params['stop_index_test']],label='test')
            plt.xlabel('time (sec)')
            axs[1].plot(range(model_params['start_index']*yaw_params['cycle_period'],model_params['stop_index']*yaw_params['cycle_period'], yaw_params['cycle_period'] ),wind_timeseries['wind_speed'][model_params['start_index']:model_params['stop_index']],label='train')
            axs[1].plot(range(model_params['start_index_test']*yaw_params['cycle_period'],model_params['stop_index_test']*yaw_params['cycle_period'],yaw_params['cycle_period'] ),wind_timeseries['wind_speed'][model_params['start_index_test']: model_params['stop_index_test']],label='test')
            plt.setp(axs[1], ylabel='wind speed (m/s)')
            plt.setp(axs[0], ylabel='wind direction (deg)')
            plt.legend()
            plt.savefig(outdir + f'{series}_results',dpi=1000)

    average_yaw_error_decrease_base = 0
    average_yaw_error_decrease_simu = 0

    if comparisons:    
        print(res_baseline_logs)
        print(res_baseline_simu)
        average_yaw_error_decrease_base = 100 * (res_baseline_logs['average yaw error_baseline_logs'] - res_model['average yaw error'])/res_baseline_logs['average yaw error_baseline_logs']
        average_yaw_error_decrease_simu = 100 * (res_baseline_simu['average yaw error_baseline_simu'] - res_model['average yaw error'])/res_baseline_simu['average yaw error_baseline_simu']

    print(res_model)

    print(f'average_yaw_error_decrease_base={average_yaw_error_decrease_base:4.2f}')
    print(f'average_yaw_error_decrease_simu={average_yaw_error_decrease_simu:4.2f}')

    energy_gain =  100*(res_model['power_control']/res_model['power_trad']-1)
    net_energy_gain = ((sum(net_prod_change) + res_model['power_trad']) / res_model['power_trad'] - 1) * 100 if net_prod_change else None

    if experiment:
        experiment.log_metric('pc_test_result', res_model['power_control'])
        experiment.log_metric('yaw_count', res_model['yaw count'])
        experiment.log_metric('mean_ye', res_model['average yaw error'])
        experiment.log_metric('energy_gain', energy_gain)
        experiment.log_metric('net_eg', net_energy_gain)
        experiment.log_metric('avg_ye_dec_base', average_yaw_error_decrease_base)
        experiment.log_metric('avg_ye_dec_simu', average_yaw_error_decrease_simu)


    return energy_gain, net_energy_gain, power_improvement, power_prod_change, conso_yaw_change, net_prod_change,rel_net_prod_change,yaw_error_rel_change



# %% ../nbs/16_environment_processing.ipynb 9
class EnvironmentProcessingFactory:
    factories = {}
    def addFactory(id, environmentProcessingFactory):
        EnvironmentProcessingFactory.factories.put[id] = environmentProcessingFactory
    addFactory = staticmethod(addFactory)
    
    # A Template Method:
    def createEnvironmentProcessing(id):
        if not EnvironmentProcessingFactory.factories.__contains__(id):
            EnvironmentProcessingFactory.factories[id] = \
              eval(id + '.Factory()')
        return EnvironmentProcessingFactory.factories[id].create()
    
    createEnvironmentProcessing = staticmethod(createEnvironmentProcessing)       


# %% ../nbs/16_environment_processing.ipynb 11
class BaseEnvironmentProcessing(ABC):
    "Base class of environment processing."
    # def __init__(self):

    def init_properties(self, args):
        self.args = args
        self.env_processing_details={}

    def set_properties(self, args):
        self.init_properties(args)

    def get_file_props(self, filepath):
        if filepath.find(sep) < 0:
            file = filepath
            property_dir = ""
        else:
            index1=filepath.rindex(sep)
            file = filepath[index1+1:]
            index2=filepath.rindex(sep, 0, index1)
            property_dir=filepath[index2+1:index1]
        drive=self.args['drive']

        return drive, property_dir, file

    # deprecated
    def add_processing_detail(self, key, value):
        self.env_processing_details[key]=value

    # deprecated
    def add_details(self, details):
        if details is not None:
            for key, value in details.items():
                self.env_processing_details[key]=value

    def add_metrics(self, metrics):
        if not metrics or len(metrics) == 0:
            return
        if not hasattr(self, 'metrics') or self.metrics is None:
            self.metrics = {}

        self.metrics.update(metrics)
 
    def add_metrics_other(self, metrics):
        if not metrics or len(metrics) == 0:
            return
        if not hasattr(self, 'metrics_other') or self.metrics_other is None:
            self.metrics_other = {}

        self.metrics_other.update(metrics)
 
    def get_metrics(self):
        return self.metrics
    
    def get_metrics_other(self):
        return self.metrics_other

    def log_properties_file(self, experiment, filepath):
        # artifact - properties file
        from comet_ml import Artifact
        index = filepath.rfind('ga-')
        print(filepath[index:])
        artifact = Artifact(filepath[index:], "Properties file")
        artifact.add(filepath)
        experiment.log_artifact(artifact)


    def get_experiment(self):
        if self.args and 'project_name' in self.args and self.args['project_name']:
            # env_name=self.args['env_name']
            # filename=self.args['file']
            # root = self.args['root_path']
            # pfile = root + self.args['configs_dir'] + env_name +sep+ filename + ".properties"
            api_key=getenv('COMET_API_KEY')
            final_ex_name = self.get_experiment_name()
            if not self.args['overwrite']:
                from comet_ml import api
                capi = api.API(api_key=api_key)
                while True:
                    connected = False
                    try:
                        experiment_exists = capi.get( workspace=self.args['workspace'] + '/' + self.args['project_name'] +'/' + final_ex_name)
                        connected = True
                    except Exception as e:
                        print ('type is:', e.__class__.__name__)
                        # print(experiment_exists)
                        print('Sleeping for 5 ...')
                        time.sleep(5)

                    if connected:
                        break

                if experiment_exists:
                    print("Experiment", final_ex_name, "already exists in", self.args['project_name'])
                    return None, True

            from comet_ml import Experiment
            experiment = Experiment(api_key=api_key,
                                    project_name=self.args['project_name'],
                                    workspace=self.args['workspace'], display_summary_level=0)

            # experiment.log_parameters(model_params)
            # experiment.log_code(path.basename(__file__))	
            experiment.set_name(final_ex_name)
        else:
            experiment = None


        return experiment, None

    def get_experiment_name(self):
        filename=self.args['file']

        prefix = filename[2:6]	
        final_ex_name = f'{prefix}-{self.args["seed"]:02}'

        return final_ex_name
    
    @abstractmethod
    def results(self, filepath=None, experiment=None, environment_properties=None, hierarchy=None):
        return None

    @abstractmethod
    def get_workspace(self):
        return 'wind-turbine'
    



# %% ../nbs/16_environment_processing.ipynb 13
class WindTurbineEnvironmentProcessing(BaseEnvironmentProcessing):
    "WindTurbine environment processing."
    # def __init__(self):

    def get_experiment_name(self):
        env_name=self.args['env_name']
        filename=self.args['file']
        root = self.args['root_path']
        pfile = root + self.args['configs_dir'] + env_name +sep+ filename + ".properties"

        ep, en = PCTRunProperties.get_environment_properties_from_filename(pfile)
        ex_name = ep['series']
        self.args['experiment_name']= ex_name
        self.args['project_name'] = self.args['project_name']+"-" + ex_name

        prefix = filename[2:6]	
        final_ex_name = f'{ex_name[0:1]}-{prefix}-{self.args["seed"]:02}'

        return final_ex_name

    def get_workspace(self):
        return 'wind-turbine'
    
    def results(self, filepath=None, experiment=None, environment_properties=None, hierarchy=None):
        plots=None
        early=None
        self.start = printtime('Start')
        
        if 'log_testing_to_experiment' in self.args:
            log_testing_to_experiment =  self.args['log_testing_to_experiment']
        else:
            log_testing_to_experiment = False


        drive, property_dir, file = self.get_file_props(filepath=filepath)

        if environment_properties is None:
            environment_properties, en = PCTRunProperties.get_environment_properties(root=drive, env='WindTurbine', property_dir=property_dir, property_file=file)
            environment_properties['keep_history'] = True
            environment_properties['range'] = 'test'
            print(environment_properties)

        if 'comparisons_print_plots' not in self.args:
            self.args['comparisons_print_plots'] = False
            print("Warning: 'comparisons_print_plots' not found in args, setting to False. Comparisons with baseline will not be computed.")

        if 'comparisons' not in self.args:
            self.args['comparisons'] = False
            print("Warning: 'comparisons' not found in args, setting to False. Comparisons with baseline will not be computed.")

        energy_gain, net_energy_gain, _ , _, _, _,_,_ = wind_turbine_results(environment_properties=environment_properties, experiment=experiment, root=drive, 
                            verbose=self.args['verbosed']['hpct_verbose'], early=early, comparisons=self.args['comparisons'], 
                            comparisons_print_plots=self.args['comparisons_print_plots'], min= not self.args['max'],
                            property_dir=property_dir, property_file=file, plots=plots, log_testing_to_experiment=log_testing_to_experiment, hierarchy=hierarchy)

        print(f'energy_gain = {energy_gain:4.2f}')
        if net_energy_gain:
            print(f'net_energy_gain = {net_energy_gain:4.2f}')

        end = printtime('End')	
        elapsed = end-self.start        
        print(f'Elapsed time: {elapsed:4.2f}')
        # NumberStats.getInstance().report()
        if experiment:
            experiment.end()

        return {'energy_gain' : energy_gain, 'net_energy_gain' : net_energy_gain}


    class Factory:
        def create(self): return WindTurbineEnvironmentProcessing()

# %% ../nbs/16_environment_processing.ipynb 15
class DummyEnvironmentProcessing(BaseEnvironmentProcessing):
    "Dummy environment processing."

    def get_workspace(self):
        return 'dummy'
    
    def results(self, filepath=None, experiment=None, environment_properties=None, hierarchy=None):

        return {}


    class Factory:
        def create(self): return DummyEnvironmentProcessing()

# %% ../nbs/16_environment_processing.ipynb 17
class GenericEnvironmentProcessing(BaseEnvironmentProcessing):
    "Generic environment processing."

    def get_workspace(self):
        return self.args['env_name']
    
    def results(self, filepath=None, experiment=None, environment_properties=None, hierarchy=None):

        return {}


    class Factory:
        def create(self): return GenericEnvironmentProcessing()

# %% ../nbs/16_environment_processing.ipynb 19
class GenericGymEnvironmentProcessing(BaseEnvironmentProcessing):
    "GenericGym environment processing."

    def set_properties(self, args):
        self.args = args
        self.env_processing_details={}
        root = self.args['root_path']
        configs_dir = self.args['configs_dir']
        env_name = self.args['env_name']
        filename = self.args['file'] + ".properties"
        pfile = root + configs_dir + env_name + sep + filename

        props = PCTRunProperties.get_properties_from_filename(pfile)
        self.args['gym_name'] = props['gym_name']

    def get_workspace(self):
        ws = self.args['gym_name']
        if 'workspace' in self.args and self.args['workspace'] != ws:
            raise ValueError(f"Workspace mismatch: existing workspace is '{self.args['workspace']}', but new workspace is '{ws}'")
        self.args['workspace'] = ws
        return ws
    
    def results(self, filepath=None, experiment=None, environment_properties=None, hierarchy=None):

        if experiment:         
            # artifact - properties file
            self.log_properties_file(experiment, filepath)

            #     hierarchy, score = PCTHierarchy.run_from_file(filepath, env_props=environment_properties, history=history, hpct_verbose= verbose, 
            # render=self.args['verbosed']['display_env'], runs=runs, experiment=experiment, min=min, plots=plots, plots_dir=self.args['plots_dir'],
            # title_prefix=title_prefix, early_termination=False
            # # ,enhanced_environment_properties=enhanced_environment_properties
            # )


            # other metrics
            experiment.log_other('LxC', self.env_processing_details['LxC'])
            experiment.log_metric('last_gen', self.env_processing_details['last_gen'])
            experiment.log_metric('score', self.env_processing_details['score'])
            experiment.log_metric('last_step', self.env_processing_details['last_step'])
            experiment.log_metrics(self.get_metrics())
            experiment.log_others(self.get_metrics_other())

        
        return {}

    class Factory:
        def create(self): return GenericGymEnvironmentProcessing()

# %% ../nbs/16_environment_processing.ipynb 21
class ARCEnvironmentProcessing(BaseEnvironmentProcessing):
    "ARC environment processing."

    def get_workspace(self):
        return 'arc-challenge'
    
    # def enhanced_environment_properties(self, environment_properties=None):
    #     enhanced_environment_properties = {}
    #     if 'dir' in environment_properties:
    #         dir = environment_properties['dir']
    #     else:
    #         dir = 'C:\\packages\\arc-prize-2024'
    #         environment_properties['dir'] = dir 

    #     if 'file_prefix' in environment_properties:
    #         file_prefix = environment_properties['file_prefix']
    #     else:
    #         file_prefix = 'arc-agi_training_'
    #         environment_properties['file_prefix'] = file_prefix

    #     file_name = path.join(dir, file_prefix) + 'challenges.json' 
    #     challenges_manager = ChallengesDataManager(file_name)
    #     data = challenges_manager.get_data_for_key(environment_properties['code'])
        
    #     self.number_of_challenges = 1
    #     if 'index' not in environment_properties:
    #         self.number_of_challenges = len(data['train'])
                
    #     enhanced_environment_properties['data']=data
    #     solutions_file = path.join(environment_properties['dir'], environment_properties['file_prefix']) + 'solutions.json' 
    #     solutions_manager = SolutionsDataManager(solutions_file)
    #     test_output_array = solutions_manager.get_data_for_key(environment_properties['code'])
    #     enhanced_environment_properties['test_output_array']=test_output_array

    #     return enhanced_environment_properties
    
    def results(self, filepath=None, experiment=None, environment_properties=None, hierarchy=None):
        print(filepath)
        
        drive, property_dir, file = self.get_file_props(filepath=filepath)
        if environment_properties is None:
            environment_properties, en = PCTRunProperties.get_environment_properties(root=drive, env='ARC', property_dir=property_dir, property_file=file)
            environment_properties['dataset'] = 'test'
            if 'index' in environment_properties:
                environment_properties['index'] = 0
            # print(environment_properties)

        # enhanced_environment_properties = self.enhanced_environment_properties(environment_properties=environment_properties)

        verbose= self.args['verbosed']['hpct_verbose'] 
        min= not self.args['max']
        history=True

        title_prefix="Test_"
        plots=self.args['hierarchy_plots'] 
        runs = int(1.5*environment_properties['runs']/DataManagerSingleton.get_instance().get_num_arrays_for_code(environment_properties['code']))
        hierarchy, score = PCTHierarchy.run_from_file(filepath, env_props=environment_properties, history=history, hpct_verbose= verbose, 
                render=self.args['verbosed']['display_env'], runs=runs, experiment=experiment, min=min, plots=plots, plots_dir=self.args['plots_dir'],
                title_prefix=title_prefix, early_termination=False
                # ,enhanced_environment_properties=enhanced_environment_properties
                )

        print(f'Test score = {score:4.3f}')
        
        fitness_list = str( [f'{i:4.3f}' for i in  self.env_processing_details['fitness_list']])
        glist = self.env_processing_details['gradient_list']
        success = hierarchy.get_environment().success(glist, self.env_processing_details['fitness'], score)    
        gradient_list = str( [f'{i:4.5f}' for i in  glist])
        print(f'Success = {success}')

        if experiment:         
            # artifact - properties file
            self.log_properties_file(experiment, filepath)

            # other metrics
            code = environment_properties['code']
            grid = hierarchy.get_grid()
            experiment.log_other('LxC', f'{len(grid)}x{max(grid)}')
            indstr = 'all'
            if 'index' in environment_properties:
                indstr = str(environment_properties['index'])
            experiment.log_other('index', indstr)
            experiment.log_other('fitness_list', fitness_list)
            experiment.log_other('gradient_list', gradient_list)
            input_set = environment_properties['input_set']
            experiment.log_other('input_set', str(input_set))
            experiment.log_metric('last_gen', self.env_processing_details['last_gen'])
            experiment.log_metric('fitness', self.env_processing_details['fitness'])
            experiment.log_metric('test_score', round(score, 3))
            experiment.log_other('code', code)
            experiment.log_other('success', success)

            # html_file
            data = DataManagerSingleton.get_instance().get_data_for_code(code)
            task_solution = DataManagerSingleton.get_instance().get_solutions_for_code(code)
            pas = PlotARCArrays()
            input_array = data['test'][0]['input']
            env_array = hierarchy.get_environment().get_env_array()
            image_file = pas.to_image("/tmp/ARC", input_array, task_solution, env_array, 'test', environment_properties['code'])
            experiment.log_image(image_file)

        return {}

    # def success(self, gradient_list, fitness, test_score):
    #     if len(gradient_list) == 0:
    #         return False        
        
    #     if fitness < self.abs_tol_ARC_resolved: # was 0.1
    #         if test_score < self.abs_tol_ARC_resolved: # was 0.5
    #             return True
            
    #     return False

    def get_experiment_name(self):
        filename=self.args['file']

        prefix = filename[3:7]	
        final_ex_name = f'{prefix}-{self.args["seed"]:02}'

        return final_ex_name


    class Factory:
        def create(self): return ARCEnvironmentProcessing()



# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/06_structure.ipynb.

# %% auto 0
__all__ = ['LevelKey', 'ArchitectureStructure']

# %% ../nbs/06_structure.ipynb 5
import enum
import uuid
from .functions import HPCTFUNCTION
from .functions import WeightedSum

# %% ../nbs/06_structure.ipynb 6
class LevelKey(enum.Enum):
   ZERO = 'level0'
   N = 'leveln'
   TOP = 'leveltop'
   ZEROTOP = 'level0top'

# %% ../nbs/06_structure.ipynb 7
class ArchitectureStructure():
    "ArchitectureStructure"
    def __init__(self, references=None, config=None, attr_mut_pb=None, lower_float=None, upper_float=None, levels_limit=None, 
                 columns_limit=None, sigma=None, mu=None, alpha=None, modes=None, namespace=None,**cargs):
        
        if namespace ==None:
            namespace = uuid.uuid1()
        self.namespace=namespace
        
        if config==None:
            self.config={'parameters': { 'modes' : {LevelKey.ZERO:3, LevelKey.N:3,LevelKey.TOP:4,LevelKey.ZEROTOP :4} }}                 
        else:
            self.config=config
        
        if modes!=None:  
            self.config['parameters']['modes']=modes


    def get_config(self):
        return self.config

    def set_node_function(self, node, function, mode, thislevel, targetlevel, targetprefix, targetcolumn, 
                          num_target_indices, inputs, input_weights, by_column, offset):
        
        func = node.get_function_from_collection(function)
        func.set_node_function(function, thislevel, targetlevel, targetprefix, targetcolumn, num_target_indices, inputs, 
                               input_weights, by_column, offset)
       
        
        #type = PCTNode.get_function_type(mode, function)
        #function_type = FunctionFactory.createFunction(type)
        #function_type.set_node_function(node, function,  thislevel, targetlevel, targetprefix, targetcolumn, num_target_indices, inputs, input_weights, by_column)
        
        
        
        
    def set_sparse_node_function(self, node, function, thislevel, input, column, input_weights):
        func = node.get_function_from_collection(function)
        func.set_sparse_node_function(function, thislevel, input, column, input_weights)

    def set_namespace(self,namespace):
        self.namespace=namespace

        
    def get_parameter(self, key):
        return self.config['parameters'][key]

    
    def set_output_function(self, node, mode, thislevel, column, input_weights):
        func = node.get_function_from_collection(HPCTFUNCTION.OUTPUT)
        func.set_output_function(thislevel, column, input_weights)

        #type = PCTNode.get_function_type(mode, HPCTFUNCTION.OUTPUT)
        #function_type = FunctionFactory.createFunction(type)
        #function_type.set_output_function(node, thislevel, column, input_weights)

    
    def set_action_function(self, hpct, env, numColumnsThisLevel,  weights):
        numActions = len(weights)
        for actionIndex in range(numActions):
            action = WeightedSum(weights=weights[actionIndex], name=f'Action{actionIndex+1}', namespace=self.namespace)
            for column in range(numColumnsThisLevel):
                action.add_link(f'OL0C{column}')
            hpct.add_postprocessor(action)
            env.add_link(action)
